import logging
import sqlite3

from foglamp.common import logger
from foglamp.common.common import _FOGLAMP_ROOT

_logger = logger.setup(__name__, level=logging.INFO)
DB_FILENAME = "management.db"
_DB_PATH = "{}/python/foglamp/services/management/data/{}".format(_FOGLAMP_ROOT, DB_FILENAME)


def create_connection(db_file):
    """ create a database connection to the SQLite database
        specified by db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
    except Exception as ex:
        _logger.exception("Error! cannot create the database connection. Exception Found: {}".format(str(ex)))
        raise
    else:
        return conn


def create_table(conn, query):
    """ create a table from the query statement
    :param conn: Connection object
    :param query: a CREATE TABLE statement
    :return:
    """
    try:
        cur = conn.cursor()
        cur.execute(query)
    except Exception as ex:
        _logger.exception("Error! cannot create the table for query: {}. Exception Found: {}".format(query, str(ex)))
        raise


def insert_table(conn, query, records=None):
    """ insert record for table from the query statement
     :param conn: Connection object
     :param query: a INSERT INTO TABLE (column_1, column_2) VALUES(?,?) statement
     :param records: insert multiple rows into table
     :return: lastrowid attribute of the Cursor object
     """
    try:
        cur = conn.cursor()
        if records is None:
            cur.execute(query)
        else:
            cur.executemany(query, records)
        conn.commit()
    except Exception as ex:
        _logger.exception("Error! cannot insert the table for query: {}. Exception Found: {}".format(query, str(ex)))
        raise
    else:
        return cur.lastrowid if records is None else cur.rowcount


def insert_json_in_table(conn, query, json_dict):
    """ insert JSON dict for table from the query statement
     :param conn: Connection object
     :param query: a INSERT INTO TABLE (column_1) VALUES(?) statement
     :param json_dict: json object
     :return: lastrowid attribute of the Cursor object
     """
    try:
        cur = conn.cursor()
        cur.execute(query, json_dict)
        conn.commit()
    except Exception as ex:
        _logger.exception("Error! cannot insert the JSON in table for query: {}. Exception Found: {}".format(query, str(ex)))
        raise
    else:
        return cur.lastrowid


def select_table(conn, query):
    """ select record for table from the query statement
         :param conn: Connection object
         :param query: a SELECT * FROM TABLE statement
         :return: fetch all records from the TABLE
    """
    try:
        cur = conn.cursor()
        cur.execute(query)
        rows = cur.fetchall()
    except Exception as ex:
        _logger.exception("Error! cannot select the table for query: {}. Exception Found: {}".format(query, str(ex)))
        raise
    else:
        return rows


def delete_from_table(conn, query):
    """ delete record for table from the query statement
         :param conn: Connection object
         :param query: a DELETE FROM TABLE statement
         :return:
    """
    try:
        cur = conn.cursor()
        cur.execute(query)
        conn.commit()
    except Exception as ex:
        _logger.exception("Error! cannot delete the records for query: {}. Exception Found: {}".format(query, str(ex)))
        raise


def update_table(conn, query):
    """ update record for table from the query statement
         :param conn: Connection object
         :param query: a UPDATE TABLE SET column_name = ? WHERE id = ?
         :return:
    """
    try:
        cur = conn.cursor()
        cur.execute(query)
        conn.commit()
    except Exception as ex:
        _logger.exception("Error! cannot update the record for query: {}. Exception Found: {}".format(query, str(ex)))
        raise


def load_database(version_entry=None):

    def _tables():
        version_table = """ CREATE TABLE IF NOT EXISTS version (
                                   id     CHAR(10)
                        );"""

        instance_version_table = """ CREATE TABLE IF NOT EXISTS instance_version (
                                                 fogman_host        character varying(80)   NOT NULL,
                                                 current            character varying(5),                     
                                                 cached             character varying(5),                           
                                                 is_synched         boolean                 NOT NULL DEFAULT 'f',
                                                 CONSTRAINT versions_pkey PRIMARY KEY (fogman_host) 
                                      ); """

        software_table = """ CREATE TABLE IF NOT EXISTS software (
                                    name              character varying(80),
                                    version           character varying(10)
                         ); """

        software_package_table = """ CREATE TABLE IF NOT EXISTS software_package (
                                            name           character varying(80)      NOT NULL,
                                            version        character varying(10)      ,
                                            CONSTRAINT     software_package_pkey      PRIMARY KEY (name) 
                                 ); """

        category_table = """ CREATE TABLE IF NOT EXISTS category (
                                    key           character varying(255)      NOT NULL,
                                    display_name  character varying(255)      NOT NULL,
                                    description   character varying(255)      NOT NULL,
                                    value         JSON                        NOT NULL DEFAULT '{}',
                                    CONSTRAINT configuration_pkey PRIMARY KEY (key)
                         ); """

        category_children_table = """ CREATE TABLE IF NOT EXISTS category_children (
                                             parent	     character varying(255)	NOT NULL,
                                             child	     character varying(255)	NOT NULL,
                                             CONSTRAINT  config_children_pkey PRIMARY KEY (parent, child)
                                  ); """

        schedule_table = """ CREATE TABLE IF NOT EXISTS schedule (
                                    process_name      character varying(255) NOT NULL,
                                    schedule_name     character varying(255) NOT NULL,
                                    schedule_type     INTEGER                NOT NULL,
                                    schedule_interval INTEGER,
                                    schedule_time     INTEGER,
                                    schedule_day      INTEGER,
                                    exclusive         boolean                NOT NULL DEFAULT 't',
                                    enabled           boolean                NOT NULL DEFAULT 'f',
                                    CONSTRAINT schedule_pkey PRIMARY KEY  (schedule_name)
                         ); """

        filter_table = """ CREATE TABLE IF NOT EXISTS filter (
                                  key           character varying(255)      NOT NULL,
                                  config        JSON                        NOT NULL DEFAULT '{}',
                                  CONSTRAINT    filter_pkey PRIMARY KEY (key)
                        ); """

        pipeline_table = """ CREATE TABLE IF NOT EXISTS pipeline (
                                    service_name  character varying(255)      NOT NULL,
                                    value         JSON                        NOT NULL DEFAULT '{}',
                                    CONSTRAINT    pipeline_pkey PRIMARY KEY (service_name)
                         ); """

        cert_store_table = """ CREATE TABLE IF NOT EXISTS certificate_store (
                                    cert         TEXT                    NOT NULL,
                                    cert_file    character varying(80),
                                    key          TEXT                    NOT NULL,
                                    key_file     character varying(80)
                         ); """

        control_script_table = """ CREATE TABLE IF NOT EXISTS control_script (
                                          name          character varying(255)        NOT NULL,
                                          steps         JSON                          NOT NULL DEFAULT '{}',
                                          acl           character varying(255),
                                          CONSTRAINT    control_script_pkey           PRIMARY KEY (name) 
                                ); """

        control_acl_table = """ CREATE TABLE IF NOT EXISTS control_acl (
                                      name          character varying(255)        NOT NULL,
                                      service       JSON                          NOT NULL DEFAULT '{}',
                                      url           JSON                          NOT NULL DEFAULT '{}',
                                      CONSTRAINT    control_acl_pkey              PRIMARY KEY (name) 
                            ); """

        attach_acl_table = """ CREATE TABLE IF NOT EXISTS attach_acl (
                                      service_name    character varying(255)        NOT NULL,
                                      acl_name        character varying(255)        NOT NULL
                            ); """

        control_pipeline_table = """CREATE TABLE IF NOT EXISTS control_pipeline (
                                           name             character  varying(255)     NOT NULL,
                                           stype            character  varying(10),       
                                           sname            character  varying(80),
                                           dtype            character  varying(10),
                                           dname            character  varying(80),
                                           enabled          boolean                     NOT NULL DEFAULT  'f',
                                           execution        character  varying(20)      NOT NULL DEFAULT  'shared',
                                           filters          character  varying(255)
                                 );"""
        buckets_table = """ CREATE TABLE IF NOT EXISTS buckets (
                                         data          JSON         NOT NULL DEFAULT  '{}'
                        );"""

        return [version_table, instance_version_table, software_table, software_package_table, category_table,
                category_children_table, schedule_table, filter_table, pipeline_table, cert_store_table,
                control_script_table, control_acl_table, attach_acl_table, control_pipeline_table, buckets_table]

    # create a database connection
    conn = create_connection(_DB_PATH)
    # create tables
    tables_list = _tables()
    with conn:
        for table in tables_list:
            create_table(conn, table)

    # insert entry in table version
    if version_entry is not None:
        insert_version_query = "INSERT INTO version (id) VALUES ({})".format(version_entry)
        with conn:
            insert_table(conn, insert_version_query)


def run_sql_file(connection, filepath):
    """ Execute Sql file
         :param connection: Connection object
         :param filepath: Sql filepath
    """
    try:
        file = open(filepath, 'r')
        sql = " ".join(file.readlines())
        cursor = connection.cursor()
        for statement in sql.split(';'):
            cursor.execute(statement)
        connection.commit()
    except Exception as ex:
        _logger.exception("Error! cannot execute Sql: {}. Exception Found: {}".format(filepath, str(ex)))
        raise
