# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

import re
import os
import uuid
import hashlib
import base64
import logging
import json
import http.client
import ssl
import socket
import time
import requests
from urllib.parse import quote

from typing import Tuple
from aiohttp import web

from foglamp.common import logger
from foglamp.services.management.common import utils
from foglamp.services.management import db
from foglamp.services.core.scheduler.entities import Schedule
from foglamp.common.common import _FOGLAMP_ROOT, _FOGLAMP_DATA

__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2022 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_logger = logger.setup(__name__, level=logging.INFO)

INBUILT_FL_SERVICES = ('storage', 'south', 'north')
INBUILT_FL_PLUGINS = ('OMF', 'DataAvailability', 'Threshold')
EXCLUDE_IN_BUILT_SOFTWARES = ["foglamp-north-OMF", "foglamp-service-south", "foglamp-service-storage",
                              "foglamp-service-north"]

def setup(app: web) -> None:
    app.router.add_route('PUT', '/poll', poll)
    # enable cors support
    utils.enable_cors(app)


def _make_https_request(host: str, action: str, uri: str, payload: dict, port: int = 5000) -> Tuple[int, dict]:
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server
    header = {'Content-type': 'application/json', 'Authorization': "Bearer {}".format(Server._access_token)}
    data = json.dumps(payload) if payload else None
    conn = http.client.HTTPSConnection(host, port, context=ssl._create_unverified_context())
    conn.request(action, uri, body=data, headers=header)
    r = conn.getresponse()
    if r.status in range(400, 500):
        _logger.error("Could not {}: {}, Client error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
    if r.status in range(500, 600):
        _logger.error("Could not {}: {}, Server error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
    res = r.read().decode()
    response = json.loads(res)
    return r.status, response


def _reconcile_local_changes(fogman_host: str, current_version: str, fl_name: str) -> None:
    """
    a) Audit Logs Stage
        send all audit logs since the last successful deployment marker was added to FogLAMP instance
    b) Bucket Data Stage
        send bucket ID and the attributes for each bucket it holds. NOTE: Do not send the bucket contents
    c) Reconcile Request
        This means FogLAMP instance has sent all data required to do a reconciliation. Now FogLAMP Manage will resolve
        any local changes as per policy and determine what needs to be sent to FogLAMP instance (in /poll endpoint)
        As of now, policies are Accept All, Reject ALl, Review
        Any response other than 200 will indicate call has failed and no reconciliation process is running on instance.
        The deployment must be aborted for the cycle.
    d) Poll Reconcile Request until its status is Completed
        Status call is made with delay of 30 seconds and with a max of 10 tries. (Tt will change as we move further...)
        No Reconciliation - no reconciliation process is in progress, it has either never been started or completed
        In Progress       - is currently executing
        Completed         - FogLAMP instance may pull the deployment info
        Failed            - deployment should be aborted for the poll cycle
        Pending Review    - is paused and manual review required thus the deployment should be aborted for the poll cycle
        Wait For Data     - Agent should send the complete content of each object listed in the response API
    """
    from foglamp.services.management.server import Server

    def send_audit():
        audit_jdoc = Server.make_http_request('GET', '/foglamp/audit?source=AUMRK&limit=1',
                                              payload=None, port=Server._core_service_port)
        if audit_jdoc:
            if 'audit' in audit_jdoc:
                aumrk_audit = audit_jdoc['audit']
                if aumrk_audit:
                    if 'timestamp' in aumrk_audit[0]:
                        timestamp = aumrk_audit[0]['timestamp'].rpartition('.')[0]
                        _logger.debug("Audit Timestamp: {}-{}".format(timestamp, quote(timestamp)))
                        audit_logs = Server.make_http_request('GET', '/foglamp/audit?since={}'.format(quote(timestamp)),
                                                              payload=None, port=Server._core_service_port)
                        res = [a for a in audit_logs['audit'] if a['source'] != 'AUMRK']
                        _logger.debug("Audit Logs to be sent: {}".format(res))
                        if res:
                            payload = {"audit": res}
                            post_audit_jdoc = _make_https_request(fogman_host, 'POST',
                                                                  '/manage/version/{}/foglamp/{}/audit'.format(
                                                                      current_version, fl_name), payload=payload)
                            _logger.debug("POST Audit API: {}".format(post_audit_jdoc))
                            if post_audit_jdoc[0] != 200:
                                raise Exception("Audit create request has failed with status code: {}".format(
                                    post_audit_jdoc[0]))

    def send_buckets():
        bucket_jdoc = Server.make_http_request('GET', '/foglamp/extension/bucket', payload=None,
                                               port=Server._core_service_port)
        _logger.debug("GET FL Buckets: {}".format(bucket_jdoc))
        if 'buckets' in bucket_jdoc:
            buckets = bucket_jdoc['buckets']
            fl_buckets = []
            for b in buckets:
                b['id'] = b['bucket']
                del b['bucket']
                del b['file']
                fl_buckets.append(b)
            payload = {"bucket": fl_buckets}
            _logger.debug("Buckets to be sent: {}".format(payload))
            post_bucket_jdoc = _make_https_request(fogman_host, 'POST',
                                                   '/manage/version/{}/foglamp/{}/bucket'.format(
                                                       current_version, fl_name), payload=payload)
            _logger.debug("POST Bucket API: {}".format(post_bucket_jdoc))
            if post_bucket_jdoc[0] != 200:
                _logger.error("Bucket create request has failed with status code: {}".format(post_bucket_jdoc[0]))

    def reconcile():
        def _get_status():
            reconcile_jdoc = _make_https_request(fogman_host,
                                                 'GET', '/manage/version/{}/foglamp/{}/reconcile'.format(
                    current_version, fl_name), payload={})
            _logger.debug("Reconcile GET Jdoc: {}".format(reconcile_jdoc))
            if reconcile_jdoc[0] == 200:
                # Status: In Progress, Completed, Failed, Pending Review, Wait For Data
                if 'status' in reconcile_jdoc[1]:
                    status_tuple = (reconcile_jdoc[1]['status'], [])
                    if 'syncData' in reconcile_jdoc[1]:
                        status_tuple = (reconcile_jdoc[1]['status'], reconcile_jdoc[1]['syncData'])
                    return status_tuple
            else:
                _logger.error("Reconcile status request has failed.")

        def _synch_data(post_payload):
            _logger.debug("POST Payload: {}".format(post_payload))
            synch_data_payload = {"syncData": {}}
            temp_cat = []
            temp_cat_children = []
            temp_sch = []
            def _add_categories(cat_payload, is_notification=False, is_services=False):
                def _add_cat_info(name, is_children=False):
                    cat_jdoc = Server.make_http_request('GET', '/foglamp/category/{}'.format(quote(name)), payload=None,
                                             port=Server._core_service_port)
                    temp_cat.append({name: cat_jdoc})
                    if is_children:
                        children_jdoc = Server.make_http_request('GET', '/foglamp/category/{}/children'.format(
                            quote(name)), payload=None, port=Server._core_service_port)
                        temp_child = []
                        for _children in children_jdoc["categories"]:
                            temp_child.append(_children["key"])
                        temp_cat_children.append({cat: temp_child})

                for cat in cat_payload:
                    if not is_services and not is_notification:
                        _add_cat_info(cat)
                    if is_services:
                        _add_cat_info(cat, is_children=True)
                        for cc in temp_cat_children:
                            if cat in cc.keys():
                                for cccp in cc[cat]:
                                    _add_cat_info(cccp)
                    if is_notification:
                        _add_cat_info(cat, is_children=True)
                        rule_cat = "rule{}".format(cat)
                        _add_cat_info(rule_cat)
                        delivery_cat = "delivery{}".format(cat)
                        _add_cat_info(delivery_cat)

            def _get_packages(services=False):
                plugins_jdoc = Server.make_http_request('GET', '/foglamp/plugins/installed', payload=None,
                                                        port=Server._core_service_port)
                synch_data_payload['syncData']['pluginsInstalled'] = plugins_jdoc["plugins"]
                if services:
                    services_jdoc = Server.make_http_request('GET', '/foglamp/service/installed', payload=None,
                                                            port=Server._core_service_port)
                    synch_data_payload['syncData']['servicesInstalled'] = services_jdoc["services"]

            def download_file_content(bid: int, buckets: list) -> dict:
                protocol = "https" if Server._is_tls_enabled else "http"
                url = '{}://{}:{}/foglamp/extension/bucket/{}'.format(protocol, Server._core_mgt_host,
                                                                   Server._core_service_port, bid)
                header = {"authorization": Server._auth_token} if Server._auth_token else {}
                response = requests.get(url, headers=header, verify=False)
                if response.status_code not in range(200, 209):
                    _logger.error("Failed to download the content of bucket id:{}".format(bid))
                    return {}
                else:
                    url_split = url.split("/bucket/")
                    filename = "/tmp/buckets/{}/{}".format(url_split[0].split("/")[-1], url_split[1])
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                    with open(filename, mode="wb") as file:
                        file.write(response.content)
                    matched_bucket = [b for b in buckets if b['bucket'] == bid]
                    final_bucket = []
                    if matched_bucket:
                        final_bucket = matched_bucket[0]
                        final_bucket['file'] = filename
                    return final_bucket

            if 'categories' in post_payload:
                _add_categories(post_payload['categories'])
            if 'datasources' in post_payload:
                _add_categories(post_payload['datasources'])
            if 'integrations' in post_payload:
                _add_categories(post_payload['integrations'])
            if 'notifications' in post_payload:
                _add_categories(post_payload['notifications'], True)
            if 'buckets' in post_payload:
                bucket_jdoc = post_payload['buckets']
                if bucket_jdoc:
                    fl_buckets = Server.make_http_request('GET', '/foglamp/extension/bucket', payload=None,
                                                          port=Server._core_service_port)
                    for _bucket in bucket_jdoc:
                        add_bucket_uri = 'https://{}:5000/manage/version/{}/bucket'.format(
                            fogman_host, current_version)
                        bucket = download_file_content(_bucket, fl_buckets['buckets'])
                        if bucket:
                            upload_bucket_along_with_association(
                                add_bucket_uri, bucket, fogman_host, current_version, fl_name,
                                {'Authorization': 'Bearer {}'.format(Server._access_token)})

            all_schedules = []
            if any(k in post_payload and post_payload[k] for k in ("schedules", "services")):
                all_sch_jdoc = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                    port=Server._core_service_port)
                all_schedules = all_sch_jdoc['schedules']

            def _add_schedule(p_sch):
                for s in p_sch:
                    for sch in all_schedules:
                        if sch['name'] == s:
                            temp_sch.append({s: sch})
                            break

            if 'services' in post_payload:
                _add_categories(post_payload['services'], is_services=True)
                _add_schedule(post_payload['services'])

            if 'schedules' in post_payload:
                _add_schedule(post_payload['schedules'])

            if 'filters' in post_payload:
                _add_categories(post_payload['filters'])

            if any(k in post_payload and post_payload[k] for k in ("notifications", "services", "filters")):
                _get_packages(services=True)

            synch_data_payload['syncData']['categories'] = temp_cat
            synch_data_payload['syncData']['categoryChildren'] = temp_cat_children
            synch_data_payload['syncData']['schedules'] = temp_sch
            _logger.debug("SYNCH DATA PAYLOAD: {}".format(json.dumps(synch_data_payload)))
            post_data = _make_https_request(fogman_host,'POST', '/manage/version/{}/foglamp/{}/syncdata'.format(
                 current_version, fl_name), payload=synch_data_payload)
            _logger.debug("Sync data POST Jdoc: {}".format(post_data))

        # Reconcile Request
        jdoc = _make_https_request(fogman_host, 'PUT', '/manage/version/{}/foglamp/{}/reconcile'.format(
            current_version, fl_name), payload={})
        _logger.debug("Reconcile Request Jdoc: {}".format(jdoc))
        if jdoc[0] == 200:
            # Poll status of reconciliation process
            tries = 1
            max_try_count = 10
            is_completed = False
            while not is_completed and tries <= max_try_count:
                status, data_sync = _get_status()
                _logger.debug("reconcile status: {}-{}".format(status, data_sync))
                if status in ('In Progress', 'Pending Review'):
                    time.sleep(30)
                    tries += 1
                    if tries == max_try_count:
                        raise Exception('Reconciliation has not completed in time. Hence Poll deployment aborted.')
                elif status == 'Wait For Data':
                    _synch_data(data_sync)
                    time.sleep(30)
                elif status == 'Completed':
                    is_completed = True
                else:
                    raise Exception('Poll deployment aborted as reconciliation process has not completed '
                                    'as got {} status.'.format(status))
        else:
            raise Exception("Reconcile request has failed with status code: {} and no reconciliation process "
                            "is running for the instance.".format(jdoc[0]))

    send_audit()
    send_buckets()
    reconcile()


def _update_process(db_conn: db, fogman_host: str, current_version: str, fl_name: str) -> None:
    def delete_tables() -> None:
        _tables = ["software", "software_package", "category", "category_children", "schedule", "filter", "pipeline",
                   "certificate_store", "control_script", "control_acl", "attach_acl", "control_pipeline", "buckets"]
        for tbl in _tables:
            db.delete_from_table(db_conn, "DELETE FROM {}".format(tbl))

    def insert_schedule_entries(data: dict) -> None:
        sch_list = []
        for sch in data:
            exclusive = 't' if sch['exclusive'] else 'f'
            enabled = 't' if sch['enabled'] else 'f'
            _type = Schedule.Type[sch['type']].value
            s_tuple = (sch['processName'], sch['name'], _type, sch['repeat'], sch['time'], sch['day'],
                       exclusive, enabled)
            sch_list.append(s_tuple)
        sch_query = "INSERT INTO schedule VALUES (?, ?, ?, ?, ?, ?, ?, ?);"
        db.insert_table(db_conn, sch_query, sch_list)

    def insert_category_entries(data: dict) -> None:
        cat_list = []
        for cat in data:
            # FIXME: It should be fix at FogMAN API side, value part should always be return for the category object
            value = cat['value'] if 'value' in cat else {}
            c_tuple = (cat['key'], cat['display_name'], cat['description'], '{}'.format(json.dumps(value)))
            cat_list.append(c_tuple)
        cat_query = "INSERT INTO category VALUES (?, ?, ?, ?);"
        db.insert_table(db_conn, cat_query, cat_list)

    def insert_category_children_entries(data: dict) -> None:
        cat_children_list = []
        for cat in data:
            for child in cat['children']:
                c_tuple = (cat['key'], child)
                cat_children_list.append(c_tuple)
        cat_child_query = "INSERT INTO category_children VALUES (?, ?);"
        db.insert_table(db_conn, cat_child_query, cat_children_list)

    def insert_software_entries(data: dict) -> None:
        software_list = []
        # TODO: Ideally below list should restrict in FogLAMP Manage API itself
        # or add optional query parameter (skip_in_built_softwares=True/False) on FogLamp API for both plugins/services
        for software in data:
            if software['component'] not in EXCLUDE_IN_BUILT_SOFTWARES:
                s_tuple = (software['component'], software['componentVersion'])
                software_list.append(s_tuple)
        sw_query = "INSERT INTO software VALUES (?, ?);"
        db.insert_table(db_conn, sw_query, software_list)

    def insert_software_package_entries(data: dict) -> None:
        software_pkg_list = []
        for software_pkg in data:
            s_tuple = (software_pkg['package'], software_pkg['version'])
            software_pkg_list.append(s_tuple)
        sw_query = "INSERT INTO software_package VALUES (?, ?);"
        db.insert_table(db_conn, sw_query, software_pkg_list)

    def insert_filter_entries(data: dict) -> None:
        filter_list = []
        for _filter in data:
            for k, v in _filter.items():
                f_tuple = (k, json.dumps(v))
                filter_list.append(f_tuple)
        filter_query = "INSERT INTO filter VALUES (?, ?);"
        db.insert_table(db_conn, filter_query, filter_list)

    def insert_pipeline_entries(data: dict) -> None:
        pipeline_list = []
        for k, v in data.items():
            p_tuple = (k, json.dumps(v))
            pipeline_list.append(p_tuple)
        pipeline_query = "INSERT INTO pipeline VALUES (?, ?);"
        db.insert_table(db_conn, pipeline_query, pipeline_list)

    def insert_certificate_store_entries(data: dict) -> None:
        cert_store_list = []
        for cs in data:
            cert_tuple = (cs['cert'], cs['cert_file'],)
            if 'key_file' in cs:
                cert_tuple = cert_tuple + (cs['key'], cs['key_file'],)
            else:
                cert_tuple = cert_tuple + ('', '',)
            cert_store_list.append(cert_tuple)
        _logger.debug("DB-CERT STORE--{}".format(cert_store_list))
        cert_store_query = "INSERT INTO certificate_store VALUES (?, ?, ?, ?);"
        db.insert_table(db_conn, cert_store_query, cert_store_list)

    def insert_control_script_entries(data: dict) -> None:
        control_script_list = []
        for cs in data:
            control_script_tuple = (cs['name'], json.dumps(cs['steps']), cs['acl'])
            control_script_list.append(control_script_tuple)
        _logger.debug("DB-CONTROL-SCRIPT--{}".format(control_script_list))
        control_script_query = "INSERT INTO control_script VALUES (?, ?, ?);"
        db.insert_table(db_conn, control_script_query, control_script_list)

    def insert_control_acl_entries(data: dict) -> None:
        control_acl_list = []
        for ca in data:
            control_acl_tuple = (ca['name'], json.dumps(ca['service']), json.dumps(ca['urls']))
            control_acl_list.append(control_acl_tuple)
        _logger.debug("DB-CONTROL-ACL--{}".format(control_acl_list))
        control_acl_query = "INSERT INTO control_acl VALUES (?, ?, ?);"
        db.insert_table(db_conn, control_acl_query, control_acl_list)

    def insert_attached_acl_entries(data: dict) -> None:
        attached_acl_list = []
        for k, v in data.items():
            aa_tuple = (k, v)
            attached_acl_list.append(aa_tuple)
        _logger.debug("DB-ATTACH-ACL--{}".format(attached_acl_list))
        attach_acl_query = "INSERT INTO attach_acl VALUES (?, ?);"
        db.insert_table(db_conn, attach_acl_query, attached_acl_list)

    def insert_control_pipeline_entries(data: dict) -> None:
        control_pipeline_list = []
        for cp in data:
            control_pipeline_tuple = (cp['name'], cp['source']['type'], cp['source']['name'],
                                      cp['destination']['type'], cp['destination']['name'], cp['enabled'],
                                      cp['execution'], '' if 'filters' not in cp else str(cp['filters']))
            control_pipeline_list.append(control_pipeline_tuple)
        _logger.debug("DB-CONTROL-PIPELINE--{}".format(control_pipeline_list))
        control_pipeline_query = "INSERT INTO control_pipeline VALUES (?, ?, ?, ?, ?, ?, ?, ?);"
        db.insert_table(db_conn, control_pipeline_query, control_pipeline_list)

    def insert_bucket_entries(data: dict) -> None:
        _logger.debug("DB-BUCKETS--{}".format(data))
        bucket_query = "INSERT INTO buckets (data) VALUES (?);"
        db.insert_json_in_table(db_conn, bucket_query, [json.dumps(data)])

    # Fetch data from FogLAMP Manage
    jdoc = _make_https_request(fogman_host, 'GET', '/manage/version/{}/foglamp/{}/data'.format(
        current_version, fl_name), payload={})
    if jdoc[0] == 200:
        with db_conn:
            # delete all entries from software, schedule, category & category_children tables
            delete_tables()
            # insert new entries
            insert_software_entries(jdoc[1]['data']['software'])
            insert_schedule_entries(jdoc[1]['data']['schedules'])
            insert_category_entries(jdoc[1]['data']['categories'])
            insert_category_children_entries(jdoc[1]['data']['categories'])
            if 'softwarePackages' in jdoc[1]['data']:
                if jdoc[1]['data']['softwarePackages']:
                    if 'packages' in jdoc[1]['data']['softwarePackages'][0]:
                        if jdoc[1]['data']['softwarePackages'][0]['packages']:
                            insert_software_package_entries(jdoc[1]['data']['softwarePackages'][0]['packages'])
            if 'filters' in jdoc[1]['data']:
                insert_filter_entries(jdoc[1]['data']['filters'])
            if 'pipelines' in jdoc[1]['data']:
                insert_pipeline_entries(jdoc[1]['data']['pipelines'])
            if 'certificateStore' in jdoc[1]['data']:
                insert_certificate_store_entries(jdoc[1]['data']['certificateStore'])
            if 'controlScripts' in jdoc[1]['data']:
                _logger.debug(jdoc[1]['data']['controlScripts'])
                insert_control_script_entries(jdoc[1]['data']['controlScripts'])
            if 'controlAcls' in jdoc[1]['data']:
                _logger.debug(jdoc[1]['data']['controlAcls'])
                insert_control_acl_entries(jdoc[1]['data']['controlAcls'])
            if 'attachedAcls' in jdoc[1]['data']:
                _logger.debug(jdoc[1]['data']['attachedAcls'])
                insert_attached_acl_entries(jdoc[1]['data']['attachedAcls'])
            if 'controlPipelines' in jdoc[1]['data']:
                _logger.debug(jdoc[1]['data']['controlPipelines'])
                insert_control_pipeline_entries(jdoc[1]['data']['controlPipelines'])
            if 'buckets' in jdoc[1]['data']:
                _logger.debug(type(jdoc[1]['data']['buckets']))
                insert_bucket_entries(jdoc[1]['data']['buckets'])
            # update cached version in instance_version table
            query = "UPDATE instance_version SET cached = '{}'".format(current_version)
            db.update_table(db_conn, query)
    else:
        _logger.warning("Failed to fetch poll data")


def _poll_data(fogman_host: str, fl_instance_name: str) -> str:
    def send_statistics_and_services_to_fogman():
        from foglamp.services.management.server import Server
        ping_info = Server.make_http_request('GET', '/foglamp/ping', payload=None, port=Server._core_service_port)
        data = {"statistics": {"uptime": ping_info["uptime"],
                               "ingress": ping_info["dataRead"],
                               "egress": ping_info["dataSent"],
                               "purged": ping_info["dataPurged"],
                               "health": ping_info["health"]}}

        data.update({"instanceVersion": ping_info["version"]})

        services = Server.make_http_request('GET', '/foglamp/service', payload=None, port=Server._core_service_port)
        svc_list = []
        for svc in services["services"]:
            svc_list.append({"name": svc["name"], "status": svc["status"]})
        data.update({"services": svc_list})
        _logger.debug('Req: /manage/version/{}/foglamp/{}/poll with data: {}'.format(current_version,
                                                                                     fl_instance_name, data))
        poll_jdoc = _make_https_request(fogman_host, 'PUT', '/manage/version/{}/foglamp/{}/poll'.format(
            current_version, fl_instance_name), payload=data)
        _logger.debug("PUT POLL JDOC---{}".format(poll_jdoc))
        if poll_jdoc[0] == 200:
            _logger.info("Statistics, version and services details pushed to FogLAMP Manage successfully.")
        else:
            _logger.error("Failed to push statistics, version and services details to FogLAMP Manage!")

    version_jdoc = _make_https_request(fogman_host, 'GET', '/manage/version', payload={})
    version = [v['version'] for v in version_jdoc[1]['versions'] if v['current']]
    current_version = version[0]
    # Always send statistics and services details to FogLAMP Manage And will be called when manage task schedule triggers
    send_statistics_and_services_to_fogman()
    db_conn = db.create_connection(db._DB_PATH)
    with db_conn:
        row = db.select_table(db_conn, "SELECT * FROM instance_version")
        if not row:
            # initial cached_version is empty
            query = "INSERT INTO instance_version (fogman_host, current, cached, is_synched) VALUES ('{}', " \
                    "'{}', '', 'f')".format(fogman_host, current_version)
            db.insert_table(db_conn, query)
            _update_process(db_conn, fogman_host, current_version, fl_instance_name)
            try:
                _synch_cached_to_foglamp(db_conn, fogman_host)
                message = "Sync complete!"
            except Exception as ex:
                _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                message = "Sync to local FogLAMP is pending."
        else:
            cached_version = row[0][2]
            synched = row[0][3]
            if current_version == cached_version:
                if synched == 't':
                    message = "Current version with FogLAMP Manage and synchronized cached version are the same. " \
                              "Hence no action required."
                    _logger.warning(message)
                else:
                    msg = "Current version with FogLAMP Manage and cached version are the same, " \
                          "but not synchronized within FogLAMP..."
                    _logger.warning(msg)
                    try:
                        _synch_cached_to_foglamp(db_conn, fogman_host)
                        message = "Sync complete!"
                    except Exception as ex:
                        _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                        message = "Sync to local FogLAMP is pending."
            else:
                query = "UPDATE instance_version SET current = '{}', is_synched = 'f'".format(current_version)
                db.update_table(db_conn, query)
                _reconcile_local_changes(fogman_host, current_version, fl_instance_name)
                _update_process(db_conn, fogman_host, current_version, fl_instance_name)
                try:
                    _synch_cached_to_foglamp(db_conn, fogman_host)
                    message = "Sync complete!"
                except Exception as ex:
                    _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                    message = "Sync to local FogLAMP is pending."
    return message


def _add_fl_and_push_data_to_fogman(fogman_host: str, fl_service_name: str) -> bool:
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server

    Server._push_state = "STARTED"
    version_jdoc = _make_https_request(fogman_host, 'GET', '/manage/version', payload={})
    # Find unlocked versions and if multiple found then sorting in alphabetical/reverse order.
    # if NOT found version list sorted by created date & in descending order and then create new version with
    # incremented value to the current version of last digit and parent_version is the current version
    # For example: current_version = 1.2.0 then new_version = 1.2.1 & parent_version = 1.2.0
    # NOTE:  Agent schedule task will try to add to new version;
    #        if FL is unknown in any existing current versions. Hence
    #        if the version in which FL was added gets locked but Not set to current i.e deployment is in progress
    versions = [v['version'] for v in version_jdoc[1]['versions'] if not v['locked']]
    if versions:
        versions.sort(reverse=True)
        version = versions[0]
    else:
        ver = version_jdoc[1]['versions']
        # Version list sorted by created date & in descending order
        sorted_version = sorted(ver, key=lambda i: i['created'], reverse=True)
        for v in sorted_version:
            if v['current']:
                base_version = v['version']
                break
            elif v['locked']:
                base_version = v['version']
            else:
                _logger.exception("Failed to find the base version to create new version")
                raise
        major, minor, patch = base_version.split('.')
        version = major + '.' + minor + '.' + str(int(patch) + 1)
        data = {"version": version, "description": "Version via management service"}
        create_version_jdoc = _make_https_request(fogman_host, 'POST', '/manage/version/{}'.format(base_version),
                                                  payload=data)
        _logger.debug("CREATE VERSION JDOC---{}".format(create_version_jdoc))

    ping_info = Server.make_http_request('GET', '/foglamp/ping', payload=None, port=Server._core_service_port)
    # Below is used to get autMethod config item details and will be used to find out username, password, cert contents
    # And send these fields to FogLAMP Manage POST API FL instance creation
    rest_api_info = Server.make_http_request('GET', '/foglamp/category/rest_api', payload=None,
                                             port=Server._core_service_port)
    data = {
        "name": fl_service_name,
        "address": Server._host,
        "servicePort": Server._core_service_port,
        "scheme": "https" if rest_api_info["enableHttp"]["value"] == "false" else "http",
        "username": "",
        "password": "",
        "certificate": "",
        "hostname": socket.gethostname(),
        "instanceVersion": ping_info["version"],
        "deployMode": "poll"
    }
    # Check authMethod only if authentication is mandatory and set username, password, certificate fields accordingly
    if not ping_info["authenticationOptional"]:
        """
            authMethod:
            any  - all empty
            pwd  - admin + admin's pwd
            cert - admin + admin's cert content
        """
        if rest_api_info["authMethod"]["value"] == "password":
            data["username"] = "admin"
            data["password"] = "foglamp"
        elif rest_api_info["authMethod"]["value"] == "certificate":
            cert_dir_path = "{}{}".format(_FOGLAMP_DATA, "/etc/certs/admin.cert") if _FOGLAMP_DATA else "{}{}".format(
                _FOGLAMP_ROOT, "/data/etc/certs/admin.cert")
            certs_dir = os.path.expanduser(cert_dir_path)
            with open(certs_dir, 'r') as f:
                cert_content = f.read()
            data["username"] = "admin"
            data["certificate"] = cert_content
    add_manual_fl_jdoc = _make_https_request(fogman_host, 'POST', '/manage/version/{}/foglamp'.format(version),
                                             payload=data)
    _logger.debug("Add FL Manual JDOC---{}".format(add_manual_fl_jdoc))

    uri = "/manage/version/{}/foglamp/{}".format(version, fl_service_name)

    def add_software():
        # Fetch FL installed services
        fl_services = Server.make_http_request('GET', '/foglamp/service/installed', payload=None,
                                               port=Server._core_service_port)
        # TODO: The version field remains empty because the FogLAMP service installed API lacks the functionality
        #  to provide the version information
        fl_services = [('foglamp-service-{}'.format(s), '', '') for s in fl_services["services"]
                       if s not in INBUILT_FL_SERVICES]
        # Fetch FL installed plugins
        jdoc = Server.make_http_request('GET', '/foglamp/plugins/installed', payload=None,
                                        port=Server._core_service_port)
        fl_plugins = [(p['packageName'], p['version'], p['name']) for p in jdoc["plugins"]
                      if p['name'] not in INBUILT_FL_PLUGINS]
        # Inbuilt plugins for the notification service, contingent upon the installation of the notification software
        if any(s[0] == 'foglamp-service-notification' for s in fl_services):
            inbuilt_plugins = [('', ping_info["version"], plugin) for plugin in INBUILT_FL_PLUGINS]
        else:
            # Case of North plugin
            inbuilt_plugins = [('', ping_info["version"], INBUILT_FL_PLUGINS[0])]
        fl_plugins.extend(inbuilt_plugins)
        # FL softwares = FL Services + FL Plugins
        fl_softwares = fl_services
        fl_softwares.extend(fl_plugins)
        _logger.debug("FL Softwares--{}".format(fl_softwares))
        for software in fl_softwares:
            _data = {"component": software[0], "componentVersion": software[1], "plugin": software[2]}
            add_software_jdoc = _make_https_request(fogman_host, 'POST', '{}/software'.format(uri), payload=_data)
            _logger.debug("Add {} Software; JDOC ---{}".format(software, add_software_jdoc))

    def add_python_package():
        get_fl_py_packages = Server.make_http_request('GET', '/foglamp/python/packages', payload=None,
                                                      port=Server._core_service_port)
        for pkg in get_fl_py_packages["packages"]:
            package_uri = "/manage/version/{}/foglamp/{}/package/python".format(version, fl_service_name)
            payload = {"package": pkg["package"], "version": pkg["version"]}
            add_py_package_jdoc = _make_https_request(fogman_host, 'POST', package_uri, payload=payload)
            _logger.debug("Add {} Python Package; JDOC ---{}".format(pkg["package"], add_py_package_jdoc))

    def add_category_and_children():
        get_fl_categories = Server.make_http_request('GET', '/foglamp/category', payload=None,
                                                     port=Server._core_service_port)
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch FL Category value by key
            value = Server.make_http_request('GET', '/foglamp/category/{}'.format(cat_name), payload=None,
                                             port=Server._core_service_port)
            _data = {"key": cat['key'],
                     "description": cat['description'],
                     "displayName": cat['displayName'],
                     "value": value
                     }
            add_category_jdoc = _make_https_request(fogman_host, 'POST', '{}/category'.format(uri), payload=_data)
            _logger.debug("Add {} Category; JDOC ---{}".format(cat_name, add_category_jdoc))

        # Add category children
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch FL Category Children
            get_fl_children = Server.make_http_request('GET', '/foglamp/category/{}/children'.format(cat_name),
                                                       payload=None, port=Server._core_service_port)
            children = get_fl_children['categories']
            if children:
                children_list = []
                for child in children:
                    children_list.append(child['key'])
                add_category_children_jdoc = _make_https_request(fogman_host, 'POST', '{}/category/{}/children'.format(
                    uri, cat_name), payload=children_list)
                _logger.debug("Add {} Category Children; JDOC ---{}".format(cat_name, add_category_children_jdoc))

    def add_schedule():
        # Fetch FL Schedules
        get_fl_schedules = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                    port=Server._core_service_port)
        fl_sch = get_fl_schedules["schedules"]
        _logger.debug("FL Schedules: {}".format(fl_sch))
        for sch in fl_sch:
            sch["type"] = Schedule.Type[sch["type"]].name
            sch["day"] = 0 if sch["day"] is None else sch["day"]
            sch["repeat"] = int(sch["repeat"])
            add_schedule_jdoc = _make_https_request(fogman_host, 'POST', '{}/schedule'.format(uri), payload=sch)
            _logger.debug("Add {} schedule; JDOC ---{}".format(sch['name'], add_schedule_jdoc))

    def add_buckets() -> None:
        get_fl_schedules = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                    port=Server._core_service_port)
        fl_bucket_schedule = [y for y in get_fl_schedules["schedules"] if (y['processName'] == 'bucket_storage_c' and
                                                                           y['enabled'] == True)]
        if fl_bucket_schedule:
            is_bucket_svc_running = is_service_running(Server, "BucketStorage")
            if is_bucket_svc_running:
                get_all_buckets = Server.make_http_request('GET', '/foglamp/extension/bucket', payload=None,
                                                           port=Server._core_service_port)

                _logger.debug("FL Buckets: {}".format(get_all_buckets))
                if 'buckets' in get_all_buckets:
                    if get_all_buckets["buckets"]:
                        for bucket in get_all_buckets["buckets"]:
                            add_bucket_uri = 'https://{}:5000/manage/version/{}/bucket'.format(fogman_host, version)
                            upload_bucket_along_with_association(
                                add_bucket_uri, bucket, fogman_host, version, fl_service_name,
                                {'Authorization': 'Bearer {}'.format(Server._access_token)})


    def add_control_scripts() -> None:
        get_fl_control_scripts = Server.make_http_request('GET', '/foglamp/control/script', payload=None,
                                                          port=Server._core_service_port)
        fl_control_scripts = get_fl_control_scripts["scripts"]
        _logger.debug("FL Control scripts: {}".format(fl_control_scripts))
        for control_script in fl_control_scripts:
            control_script_uri = "/manage/version/{}/control/script".format(version)
            add_control_script_jdoc = _make_https_request(fogman_host, 'POST', control_script_uri,
                                                          payload=control_script)
            _logger.debug("Add control script with name {}; JDOC {}".format(
                control_script['name'], add_control_script_jdoc))
            if add_control_script_jdoc[0] != 200:
                _logger.error("Add {} control script failed.".format(control_script))

    def add_control_acls() -> None:
        get_fl_control_acls = Server.make_http_request('GET', '/foglamp/ACL', payload=None,
                                                       port=Server._core_service_port)
        fl_control_acls = get_fl_control_acls["acls"]
        _logger.debug("FL Control ACLs: {}".format(fl_control_acls))
        for control_acl in fl_control_acls:
            control_acl_uri = "/manage/version/{}/acl".format(version)
            control_acl['urls'] = control_acl['url']
            add_control_acl_jdoc = _make_https_request(fogman_host, 'POST', control_acl_uri, payload=control_acl)
            _logger.debug("Add control ACL with name {}; JDOC {}".format(
                control_acl['name'], add_control_acl_jdoc))

    def add_control_pipelines() -> None:
        get_fl_control_pipelines = Server.make_http_request('GET', '/foglamp/control/pipeline', payload=None,
                                                            port=Server._core_service_port)
        fl_control_pipelines = get_fl_control_pipelines["pipelines"]
        _logger.debug("FL Control Pipelines: {}".format(fl_control_pipelines))
        for cp in fl_control_pipelines:
            control_pipeline_uri = "/manage/version/{}/control/pipeline".format(version)
            add_control_pipeline_jdoc = _make_https_request(fogman_host, 'POST', control_pipeline_uri, payload=cp)
            _logger.debug("Add control Pipeline with name {}; JDOC {}".format(cp['name'],
                                                                              add_control_pipeline_jdoc))
            # TODO: FOGL-8101 support filters

    add_software()
    add_python_package()
    add_category_and_children()
    add_schedule()
    add_buckets()
    add_control_acls()
    add_control_scripts()
    add_control_pipelines()

    # update state to Managed for instance added in POLL mode after initial sync is completed at FogLAMP Manage
    data = {"syncStatus": "complete"}
    _logger.debug('/manage/version/{}/foglamp/{}/poll'.format(version, fl_service_name))
    put_state_jdoc = _make_https_request(fogman_host, 'PUT', '/manage/version/{}/foglamp/{}/poll'.format(
        version, fl_service_name), payload=data)
    _logger.debug("PUT FL update state JDOC---{}".format(put_state_jdoc))

    # TODO: Handle ABORTED, FAILED state
    Server._push_state = "COMPLETED"
    return True  # TODO: Server._push_state


def _synch_cached_to_foglamp(db_conn: db, fm_host: str):
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server
    roll_back_items = []
    """ Roll back item list will be used if any API call is failed over FL instance """
    MAX_TRIES = 10
    """ Max try count to hit GET package status """
    WAIT_TIME = 15
    """ Sleep/wait time on every GET package status retry """

    # Fetch categories & schedules from FL API
    get_fl_categories = Server.make_http_request('GET', '/foglamp/category', payload=None,
                                                 port=Server._core_service_port)
    get_fl_schedules = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                port=Server._core_service_port)

    def get_diff(old: dict, new: dict) -> list:
        diff = list()
        for key in new:
            if key in old:
                if old[key] != new[key]:
                    diff.append(key)
            else:
                diff.append(key)
        return diff

    def fl_categories() -> list:
        cat_list = []
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch Category value by name from FL API; as its needed to find out the diff with cached schedules
            # And required only for Update action on FL side
            value = Server.make_http_request('GET', '/foglamp/category/{}'.format(cat_name), payload=None,
                                             port=Server._core_service_port)
            fl_temp_dict = {"key": cat['key'],
                            "description": cat['description'],
                            "displayName": cat['displayName'],
                            "value": value
                            }
            cat_list.append(fl_temp_dict)
        return cat_list

    def cached_schedules() -> list:
        # Fetch Cached Schedules
        get_cached_schedules = db.select_table(db_conn, "SELECT * FROM schedule")
        sch_list = []
        # Conversions required for cached schedules; to get actual diff with fl schedules (API vs DB result)
        for sch in get_cached_schedules:
            temp_dict = {
                "processName": sch[0],
                "name": sch[1],
                "type": Schedule.Type(sch[2]).name,
                "repeat": sch[3],
                "time": sch[4],
                "day": sch[5] if sch[5] else None,
                "exclusive": True if sch[6] == 't' else False,
                "enabled": True if sch[7] == 't' else False
            }
            sch_list.append(temp_dict)
        return sch_list

    def cached_categories() -> list:
        # Fetch Cached Categories
        get_cached_categories = db.select_table(db_conn, "SELECT * FROM category")
        cat_list = []
        for cat in get_cached_categories:
            temp_dict = {"key": cat[0],
                         "description": cat[1],
                         "displayName": cat[2],
                         "value": cat[3]
                         }
            cat_list.append(temp_dict)
        return cat_list

    def cached_cert_stores() -> list:
        # Fetch Cached Cert stores
        get_cached_cert_stores = db.select_table(db_conn, "SELECT * FROM certificate_store")
        cached_cert_store_list = []
        for cert_store in get_cached_cert_stores:
            cert_tuple = (cert_store[0], cert_store[1], cert_store[2], cert_store[3])
            cached_cert_store_list.append(cert_tuple)
        return cached_cert_store_list

    def cached_scripts() -> list:
        # Fetch Cached Control scripts
        get_cached_scripts = db.select_table(db_conn, "SELECT * FROM control_script")
        cached_script_list = []
        for control_script in get_cached_scripts:
            temp_dict = {"name": control_script[0], "steps": control_script[1], "acl": control_script[2]}
            cached_script_list.append(temp_dict)
        return cached_script_list

    def cached_acls() -> list:
        # Fetch Cached Control ACLs
        get_cached_acls = db.select_table(db_conn, "SELECT * FROM control_acl")
        cached_acl_list = []
        for control_acl in get_cached_acls:
            temp_dict = {"name": control_acl[0], "service": control_acl[1], "url": control_acl[2]}
            cached_acl_list.append(temp_dict)
        return cached_acl_list

    def cached_control_pipelines() -> list:
        destination_lookup = _make_https_request(fm_host, 'GET', '/foglamp/control/lookup?type=destination',
                                                 payload={})
        source_lookup = _make_https_request(fm_host, 'GET', '/foglamp/control/lookup?type=source',
                                            payload={})
        _logger.debug(destination_lookup)
        _logger.debug(source_lookup)

        def get_type_id(lookup_type, lookup_value):
            if lookup_type == 'destination':
                convert_dtype = [dl['cpdid'] for dl in destination_lookup[1] if dl['name'] == lookup_value]
                return convert_dtype[0] if convert_dtype else 1
            else:
                convert_stype = [sl['cpsid'] for sl in source_lookup[1] if sl['name'] == lookup_value]
                return convert_stype[0] if convert_stype else 1

        # Fetch Cached Control Pipelines
        get_cached_control_pipelines = db.select_table(db_conn, "SELECT * FROM control_pipeline")
        cached_cp_list = []
        for cp in get_cached_control_pipelines:
            temp_dict = {
                "name": cp[0],
                "stype": get_type_id('source', cp[1]), "sname": cp[2],
                "dtype": get_type_id('destination', cp[3]), "dname": cp[4],
                "enabled": True if cp[5] else False,
                "execution": cp[6],
                "filters": cp[7]
            }
            cached_cp_list.append(temp_dict)
        return cached_cp_list

    def cached_buckets() -> list:
        get_cached_buckets = db.select_table(db_conn, "SELECT * FROM buckets")
        cached_buckets_list = []
        for _bucket in get_cached_buckets:
            temp_dict = {"data": _bucket}
            cached_buckets_list.append(temp_dict)
        return cached_buckets_list

    fl_cat = fl_categories()
    _logger.debug("FL Categories: {}".format(fl_cat))
    fl_sch = get_fl_schedules["schedules"]
    _logger.debug("FL Schedules: {}".format(fl_sch))
    cached_cat = cached_categories()
    _logger.debug("Cached Categories: {}".format(cached_cat))
    cached_sch = cached_schedules()
    _logger.debug("Cached Schedules: {}".format(cached_sch))
    get_cached_filters = db.select_table(db_conn, "SELECT * FROM filter")
    _logger.debug("Cached Filters: {}".format(get_cached_filters))
    get_cached_pipelines = db.select_table(db_conn, "SELECT * FROM pipeline")
    _logger.debug("Cached Pipelines: {}".format(get_cached_pipelines))
    get_fl_certificate_stores = Server.make_http_request('GET', '/foglamp/certificate', payload=None,
                                                         port=Server._core_service_port)
    fl_certs = get_fl_certificate_stores['certs'] + get_fl_certificate_stores['keys']
    fl_inbuilt_certs = ['user.cert', 'ca.cert', 'user.csr', 'foglamp.cert', 'admin.cert', 'admin.csr', 'foglamp.csr',
                        'admin.key', 'ca.key', 'foglamp.key', 'user.key']
    fl_certs_without_inbuilt = [c for c in fl_certs if c not in fl_inbuilt_certs]
    _logger.debug("FL Certificate stores list: {}".format(fl_certs_without_inbuilt))
    cached_certs = cached_cert_stores()
    _logger.debug("Cached Certificate stores: {}".format(cached_certs))
    cached_cert_store_for_deletion = []
    get_fl_control_scripts = Server.make_http_request('GET', '/foglamp/control/script', payload=None,
                                                      port=Server._core_service_port)
    fl_control_scripts = get_fl_control_scripts["scripts"]
    _logger.debug("FL Control scripts: {}".format(fl_control_scripts))
    cached_control_scripts = cached_scripts()
    _logger.debug("Cached Control scripts: {}".format(cached_control_scripts))
    get_fl_control_acls = Server.make_http_request('GET', '/foglamp/ACL', payload=None,
                                                   port=Server._core_service_port)
    fl_control_acls = get_fl_control_acls["acls"]
    _logger.debug("FL Control ACLs: {}".format(fl_control_acls))
    cached_control_acls = cached_acls()
    _logger.debug("Cached Control ACLs: {}".format(cached_control_acls))
    get_fl_control_pipelines = Server.make_http_request('GET', '/foglamp/control/pipeline', payload=None,
                                                        port=Server._core_service_port)
    fl_control_pipelines = get_fl_control_pipelines["pipelines"]
    _logger.debug("FL Control Pipelines: {}".format(fl_control_pipelines))
    cached_cp = cached_control_pipelines()
    _logger.debug("Cached Control Pipelines: {}".format(cached_cp))

    # TODO: First check whether bucket is installed or not and should be in running state
    get_fl_buckets = Server.make_http_request('GET', '/foglamp/extension/bucket', payload=None,
                                              port=Server._core_service_port)
    _logger.debug("GET - FL buckets...{}".format(get_fl_buckets))
    cached_buckets_data = cached_buckets()
    _logger.debug("Cached Buckets: {}".format(cached_buckets_data))

    # Get version and FogLAMP instance name
    instance_version = db.select_table(db_conn, "SELECT current FROM instance_version")
    fl_current_version = instance_version[0][0]
    svc_name = [y for y in fl_cat if y['key'] == 'service']
    fl_svc_name = svc_name[0]['value']['name']['value'] if svc_name else 'FogLAMP'

    def install_software():
        softwares = db.select_table(db_conn, "SELECT * FROM software")

        # Fetch Services from FL API
        fl_services = Server.make_http_request('GET', '/foglamp/service/installed', payload=None,
                                               port=Server._core_service_port)
        fl_services = ['foglamp-service-{}'.format(s) for s in fl_services["services"] if s not in INBUILT_FL_SERVICES]
        _logger.debug("FL Services--{}".format(fl_services))
        # Fetch Cached Services
        cached_services = [(n, v) for n, v in softwares if
                           n not in fl_services and str(n).startswith('foglamp-service-')]
        _logger.debug("Cached Services--{}".format(cached_services))

        svc_payloads = [{"format": "repository", "name": n, "version": v} for n, v in cached_services]
        for data in svc_payloads:
            jdoc = Server.make_http_request('POST', '/foglamp/service?action=install',
                                            payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-SERVICE INSTALLATION: {}".format(jdoc))
            max_retry_count = MAX_TRIES
            while max_retry_count:
                # GET Package Status
                package_status_jdoc = Server.make_http_request('GET', '/{}'.format(jdoc["statusLink"]), payload=data,
                                                               port=Server._core_service_port)
                _logger.debug("JDOC - GET-SERVICE-PACKAGE STATUS: {} on retry count {}".format(package_status_jdoc,
                                                                                               max_retry_count))
                if package_status_jdoc['packageStatus'][0]['status'] == "success":
                    break
                elif package_status_jdoc['packageStatus'][0]['status'] == "failed":
                    # TODO: FOGL-4503 (Low priority) for Software Rollback
                    # And even not required as it is unnecessary overhead
                    raise Exception('{} package installation failed'.format(data['name']))
                # sleep time added b/w retries
                time.sleep(WAIT_TIME)
                max_retry_count -= 1

        # Fetch Plugins from FL API
        jdoc = Server.make_http_request('GET', '/foglamp/plugins/installed', payload=None,
                                        port=Server._core_service_port)
        fl_plugins = [p['packageName'] for p in jdoc["plugins"] if p['name'] not in INBUILT_FL_PLUGINS]
        _logger.debug("FL Plugins--{}".format(fl_plugins))

        prefix_plugins = ['foglamp-south-', 'foglamp-north-', 'foglamp-filter-', 'foglamp-notify-', 'foglamp-rule-']
        # Fetch Cached Plugins
        cached_plugins = [(n, v) for n, v in softwares if n not in fl_plugins and str(n).startswith(
            tuple(prefix_plugins))]
        _logger.debug("Cached Plugins---{}".format(cached_plugins))
        plugin_payloads = [{"format": "repository", "name": n, "version": v} for n, v in cached_plugins]
        for data in plugin_payloads:
            jdoc = Server.make_http_request('POST', '/foglamp/plugins', payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-PLUGINS INSTALL: {}".format(jdoc))
            max_retry_count = MAX_TRIES
            while max_retry_count:
                # GET Package Status
                package_status_jdoc = Server.make_http_request('GET', '/{}'.format(jdoc["statusLink"]), payload=data,
                                                               port=Server._core_service_port)
                _logger.debug("JDOC - GET-PACKAGE STATUS: {} on retry count {}".format(package_status_jdoc,
                                                                                       max_retry_count))
                if package_status_jdoc['packageStatus'][0]['status'] == "success":
                    break
                elif package_status_jdoc['packageStatus'][0]['status'] == "failed":
                    # TODO: FOGL-4503 (Low priority) for Software Rollback
                    # And even not required as it is unnecessary overhead
                    raise Exception('{} package installation failed'.format(data['name']))
                # sleep time added b/w retries
                time.sleep(WAIT_TIME)
                max_retry_count -= 1

    def install_python_packages():
        software_package = db.select_table(db_conn, "SELECT * FROM software_package")
        cached_python_packages = [n for n, v in software_package]
        get_fl_py_packages = Server.make_http_request('GET', '/foglamp/python/packages', payload=None,
                                                      port=Server._core_service_port)
        fl_python_packages = [p['package'] for p in get_fl_py_packages["packages"]]
        list_diff = list(set(cached_python_packages) - set(fl_python_packages))
        _logger.debug("DIFF of Python Package list: {}".format(list_diff))
        # NOTE: DELETE route is not available at FogLAMP side
        for ld in list_diff:
            payload = {"package": ld}
            version = [v for n, v in software_package if ld == n]
            if version:
                payload["version"] = version[0]
            add_python_pkg_jdoc = Server.make_http_request('POST', '/foglamp/python/package',
                                                           payload=payload, port=Server._core_service_port)
            _logger.debug("POST Python package Payload: {};  JDOC: {}".format(payload, add_python_pkg_jdoc))

    def download_file_content(bucket: dict, version: str) -> dict:
        url = 'https://{}:5000/manage/version/{}/bucket/{}'.format(fm_host, version, bucket['bucket'])
        response = requests.get(url, headers={'Authorization': "Bearer {}".format(Server._access_token)}, verify=False)
        if response.status_code not in range(200, 209):
            _logger.error("Failed to download the content of bucket id:{}".format(bucket['bucket']))
            return {}
        else:
            url_split = url.split("/bucket/")
            filename = "/tmp/buckets/{}/{}".format(url_split[0].split("/")[-1], url_split[1])
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, mode="wb") as file:
                file.write(response.content)
            bucket['file'] = filename
            return bucket

    def update_buckets() -> None:
        def add_update_delete_buckets():
            non_deleted_buckets = []
            # Update case
            for cb in cached_buckets_data:
                if 'data' in cb:
                    data = json.loads(cb['data'][0])
                    if 'update' in data:
                        _logger.debug("Buckets UPDATE list: {}".format(data['update']))
                        for u in data['update']:
                            attrs = {"attributes": u['attributes']}
                            jdoc = Server.make_http_request('PUT', '/foglamp/extension/bucket/{}'.format(
                                u['foglamp_bucket']), payload=attrs, port=Server._core_service_port)
                            _logger.debug("PUT- BUCKET ATTRS for id:{} - JDOC: {}".format(u['bucket'], jdoc))
                            non_deleted_buckets.append(u['foglamp_bucket'])
            # Delete case
            if 'message' not in get_fl_buckets:
                for x in get_fl_buckets['buckets']:
                    if x['bucket'] not in non_deleted_buckets:
                        jdoc = Server.make_http_request('DELETE', '/foglamp/extension/bucket/{}'.format(
                            x['bucket']), payload=None, port=Server._core_service_port)
                        _logger.debug("DELETE- BUCKET id:{} - JDOC: {}".format(x['bucket'], jdoc))
            # Add case
            for cb in cached_buckets_data:
                if 'data' in cb:
                    data = json.loads(cb['data'][0])
                    if 'add' in data:
                        _logger.debug("Buckets Create list: {}".format(data['add']))
                        for a in data['add']:
                            bucket_info = download_file_content(a, fl_current_version)
                            if bucket_info:
                                protocol = "https" if Server._is_tls_enabled else "http"
                                add_bucket_uri = '{}://{}:{}/foglamp/extension/bucket'.format(
                                    protocol, Server._core_mgt_host, Server._core_service_port)
                                header = {"authorization": Server._auth_token} if Server._auth_token else {}
                                upload_bucket_along_with_association(
                                    add_bucket_uri, bucket_info, fm_host, fl_current_version, fl_svc_name, header)


        """ Update buckets:
            a) If schedule exists at FogLAMP instance then case is to add its schedule and wait until it's up and 
               also remove its entry from cached schedules (i.e deployed poll data) to avoid other attempt to add
               its schedule again.
            b) If schedule does not exist at FogLAMP instance; there will be NO Action as unlike other schedules
               update properties of schedule and if enabled property is changed then enable/disable calls
               There is no chance to update the bucket schedule from FogLAMP Manage side 
        """
        if cached_buckets_data:
            fl_bucket_schedule = [y for y in fl_sch if y['processName'] == 'bucket_storage_c']
            if not fl_bucket_schedule:
                bucket_schedule = [x for x in cached_sch if x['processName'] == 'bucket_storage_c']
                if bucket_schedule:
                    temp = bucket_schedule[0]
                    temp["type"] = Schedule.Type[temp["type"]].value
                    temp["day"] = 0 if temp["day"] is None else temp["day"]
                    temp["process_name"] = temp["processName"]
                    del temp["processName"]
                    data = temp
                    jdoc = Server.make_http_request('POST', '/foglamp/schedule', payload=data,
                                                    port=Server._core_service_port)
                    _logger.debug("JDOC - Bucket POST-SCHEDULE: {}".format(jdoc))
                    if 'schedule' in jdoc:
                        # Remove bucket schedule from cached schedules as same as deployed data
                        cached_sch.remove(temp)
                        is_bucket_svc_running = is_service_running(Server, "BucketStorage")
                        if is_bucket_svc_running:
                            add_update_delete_buckets()
            else:
                add_update_delete_buckets()

    def add_and_update_category():
        # Find Notification parent category
        parent_notification_cat_name = 'Notifications'
        is_notification_parent_category_found = False
        for flk in fl_cat:
            if flk['key'] == parent_notification_cat_name:
                is_notification_parent_category_found = True
        _logger.debug("Notification Parent category found: {}".format(is_notification_parent_category_found))

        # Case of update category:
        for x in fl_cat:
            for y in cached_cat:
                if x['key'] == y['key']:
                    # For matched category; Find diff of cached & FL category and if found then update on FL side
                    fl_cat_list = []
                    cached_cat_list = []
                    for k1, v1 in x['value'].items():
                        fl_cat_list.append({k1: v1['value']})
                    for k2, v2 in json.loads(y['value']).items():
                        cached_cat_list.append({k2: v2['value']})
                    data = {}
                    for config_item_with_value in cached_cat_list:
                        if config_item_with_value not in fl_cat_list:
                            data.update(config_item_with_value)
                    if data:
                        _logger.debug("UPDATE {} category data: {}".format(x['key'], data))
                        # SPECIAL CASE:
                        # if notification parent is found then fetch its children and also fetch notification instances
                        # then, check requested category in available notification instances and having change in rule,
                        # channel config items value
                        # then, DELETE that notification instance and recreate instance with new configuration
                        # then, return immediately
                        if is_notification_parent_category_found:
                            # Fetch parent notifications children
                            notification_children_jdoc = Server.make_http_request(
                                'GET', '/foglamp/category/{}/children'.format(parent_notification_cat_name),
                                payload=None, port=Server._core_service_port)
                            _logger.debug("GET-NOTIFICATION-CHILDREN - JDOC: {}".format(
                                notification_children_jdoc))
                            notification_children = [nc['key'] for nc in notification_children_jdoc['categories']]
                            # Fetch notification instances list
                            notification_instances_jdoc = Server.make_http_request(
                                'GET', '/foglamp/notification'.format(parent_notification_cat_name), payload=None,
                                port=Server._core_service_port)
                            _logger.debug("GET NOTIFICATION INSTANCES - JDOC: {}".format(
                                notification_instances_jdoc))
                            notification_instances = [nc['name'] for nc in notification_instances_jdoc['notifications']]
                            # check for instance if either rule or channel config item value has changed
                            if x['key'] in notification_children and x['key'] in notification_instances:
                                if any(cfg_item in data for cfg_item in ('rule', 'channel')):
                                    _logger.info("Removing {} notification instance as either rule or channel "
                                                 "value has been changed".format(x['key']))
                                    notification_instance_delete_jdoc = Server.make_http_request(
                                        'DELETE', '/foglamp/notification/{}'.format(x['key']), payload=None,
                                        port=Server._core_service_port)
                                    _logger.debug("DELETE {} NOTIFICATION INSTANCE - JDOC: {}".format(
                                        x['key'], notification_instance_delete_jdoc))
                                    # Extra wait time is required so that related rule & delivery gets deleted
                                    time.sleep(WAIT_TIME - 5)
                                    if 'result' in notification_instance_delete_jdoc:
                                        _logger.info(notification_instance_delete_jdoc['result'])
                                        # Recreate instance again with new configuration
                                        # rule_config items payload
                                        rule_config = {}
                                        for rule_category in cached_cat:
                                            if rule_category['key'] == 'rule{}'.format(x['key']):
                                                for r_key, r_val in json.loads(rule_category['value']).items():
                                                    rule_config[r_key] = r_val['value']
                                                break
                                        _logger.debug("rule_config items payload: {}".format(rule_config))
                                        # delivery_config items payload
                                        delivery_config = {}
                                        for delivery_category in cached_cat:
                                            if delivery_category['key'] == 'delivery{}'.format(x['key']):
                                                for d_key, d_val in json.loads(delivery_category['value']).items():
                                                    delivery_config[d_key] = d_val['value']
                                                break
                                        _logger.debug("delivery_config items payload: {}".format(delivery_config))
                                        jv = json.loads(y['value'])
                                        post_notification_payload = {"name": y['key'],
                                                                     "description": y['description'],
                                                                     "rule": jv['rule']['value'],
                                                                     "channel": jv['channel']['value'],
                                                                     "notification_type": jv['notification_type'][
                                                                         'value'],
                                                                     "enabled": jv['enable']['value'],
                                                                     "rule_config": rule_config,
                                                                     "delivery_config": delivery_config}
                                        _logger.debug("{} NOTIFICATION PAYLOAD: {}".format(x['key'],
                                                                                           post_notification_payload))
                                        new_notification_instance_jdoc = Server.make_http_request(
                                            'POST', '/foglamp/notification', payload=post_notification_payload,
                                            port=Server._core_service_port)
                                        _logger.debug("POST NOTIFICATION INSTANCE JDOC: {}".format(
                                            new_notification_instance_jdoc))
                                        if 'result' in new_notification_instance_jdoc:
                                            _logger.info(new_notification_instance_jdoc['result'])
                                        else:
                                            _logger.error("{} notification instance recreation failed".format(x['key']))
                                    else:
                                        _logger.error("{} notification instance deletion failed".format(x['key']))
                                    return
                        # SPECIAL CASE:
                        # if plugin key is found in category object & different then check schedule with the same category name
                        # if exists then Call delete service or task endpoint on the basis of processName
                        # then recreate it
                        if 'plugin' in data:
                            # TODO: FOGL-5542
                            # if storage category plugin change will be detected, we shall just set a restart flag
                            #  and restart when Poll agent sync is completed
                            _logger.info("{} plugin change detection case for {} category".format(data['plugin'],
                                                                                                  x['key']))
                            _logger.debug("For {} cat, plugin value changed to {}".format(x['key'], data['plugin']))
                            sch_info = [(c['name'], c['processName']) for c in fl_sch if c['name'] == x['key']]
                            if sch_info:
                                def _reuse_category_config():
                                    reuse_cat_info = [cc for cc in cached_cat if x['key'] == cc['key']]
                                    _logger.debug("New cat config: {}".format(reuse_cat_info))
                                    reuse_cat_config = {}
                                    for nk, nv in json.loads(reuse_cat_info[0]['value']).items():
                                        reuse_cat_config[nk] = nv
                                    _config = [(rcc, reuse_cat_config[rcc]['value'])
                                               for rcc in reuse_cat_config if 'plugin' not in rcc]
                                    value_config = {}
                                    for kc, vc in _config:
                                        value_config.update({kc: {"value": vc}})
                                    return reuse_cat_info, reuse_cat_config, value_config

                                if sch_info[0][1] in ['south_c', 'north_C']:
                                    _logger.debug("DELETE foglamp/service/{}".format(quote(x['key'])))
                                    delete_svc_jdoc = Server.make_http_request('DELETE', '/foglamp/service/{}'.format(
                                        quote(x['key'])), payload=None, port=Server._core_service_port)
                                    _logger.debug("JDOC - DELETE-SERVICE: {}".format(delete_svc_jdoc))
                                    # 200-Case
                                    if 'result' in delete_svc_jdoc:
                                        # Recreate service
                                        reuse_sch_info = [csh for csh in cached_sch if csh["name"] == x['key']]
                                        _logger.debug("Reused cached info: {}".format(reuse_sch_info))
                                        if reuse_sch_info and reuse_sch_info[0] in cached_sch:
                                            _new_cat_info, _new_cat_config, _new_val_config = _reuse_category_config()
                                            _logger.debug("Add SERVICE POST foglamp/service/{}".format(quote(x['key'])))
                                            svc_payload = {"name": _new_cat_info[0]['key'],
                                                           "plugin": _new_cat_config['plugin']['value'],
                                                           "type": "south" if sch_info[0][1] == "south_c" else "north",
                                                           "enabled": reuse_sch_info[0]['enabled'],
                                                           "config": _new_val_config
                                                           }
                                            _logger.debug("{} new Service with payload: {}".format(
                                                quote(x['key']), json.dumps(svc_payload)))
                                            # Recreate service
                                            new_svc_jdoc = Server.make_http_request('POST', '/foglamp/service',
                                                                                    payload=svc_payload,
                                                                                    port=Server._core_service_port)
                                            _logger.debug("JDOC - POST-SERVICE: {}".format(new_svc_jdoc))
                                            if 'name' in new_svc_jdoc:
                                                if x['key'] == new_svc_jdoc['name']:
                                                    _logger.info("Successfully recreated {} service".format(x['key']))
                                                    # Remove schedule entry from schedules list
                                                    cached_sch.remove(reuse_sch_info[0])
                                                    for fl_index, fl_val in enumerate(fl_sch):
                                                        if fl_val['name'] == x['key']:
                                                            fl_sch.pop(fl_index)
                                            else:
                                                _logger.error("{} schedule not found in poll agent database".format(
                                                    x['key']))
                                        else:
                                            _logger.error("{} service recreation failed".format(x['key']))
                                    else:
                                        _logger.error("DELETE {} service failed".format(quote(x['key'])))
                                elif sch_info[0][1] in ['north', 'north_c']:
                                    _logger.debug("DELETE: foglamp/scheduled/task/{}".format(quote(x['key'])))
                                    del_task_jdoc = Server.make_http_request('DELETE',
                                                                             '/foglamp/scheduled/task/{}'.format(
                                                                                 quote(x['key'])), payload=None,
                                                                             port=Server._core_service_port)
                                    _logger.debug("JDOC - DELETE-TASK: {}".format(del_task_jdoc))
                                    # 200-Case
                                    if 'result' in del_task_jdoc:
                                        # Recreate task
                                        reuse_sch_info = [csh for csh in cached_sch if csh["name"] == x['key']]
                                        _logger.debug("Reused cached info: {}".format(reuse_sch_info))
                                        if reuse_sch_info and reuse_sch_info[0] in cached_sch:
                                            _new_cat_info, _new_cat_config, _new_val_config = _reuse_category_config()
                                            task_payload = {"name": _new_cat_info[0]['key'],
                                                            "plugin": _new_cat_config['plugin']['value'],
                                                            "type": "north",
                                                            "schedule_repeat": reuse_sch_info[0]['repeat'],
                                                            "schedule_type": "3",  # here INTERVAL type always
                                                            "schedule_enabled": reuse_sch_info[0]['enabled'],
                                                            "config": _new_val_config
                                                            }
                                            _logger.debug("{} new task foglamp/scheduled/task with payload: {}".format(
                                                quote(x['key']), json.dumps(task_payload)))
                                            new_task_jdoc = Server.make_http_request('POST', '/foglamp/scheduled/task',
                                                                                     payload=task_payload,
                                                                                     port=Server._core_service_port)
                                            _logger.debug("JDOC - POST-TASK: {}".format(new_task_jdoc))
                                            if 'name' in new_task_jdoc:
                                                if x['key'] == new_task_jdoc['name']:
                                                    _logger.info("Successfully recreated {} task".format(x['key']))
                                                    # Remove schedule entry from schedules list
                                                    cached_sch.remove(reuse_sch_info[0])
                                                    for fl_index, fl_val in enumerate(fl_sch):
                                                        if fl_val['name'] == reuse_sch_info[0]['name']:
                                                            fl_sch.pop(fl_index)
                                            else:
                                                _logger.error("{} task recreation failed".format(x['key']))
                                        else:
                                            _logger.error("{} schedule not found in poll agent database".format(
                                                x['key']))
                                    else:
                                        _logger.error("DELETE {} task failed".format(quote(x['key'])))
                                else:
                                    # BAD case - Nothing to do
                                    logger.debug("Nothing to delete as processName not in "
                                                 "['south_c', 'north_C', 'north', 'north_c']")
                            else:
                                # Notification related categories
                                _logger.warning("No such action is required for {} notification category".format(
                                    y['key']))
                        else:
                            jdoc = Server.make_http_request('PUT', '/foglamp/category/{}'.format(quote(y['key'])),
                                                            payload=data, port=Server._core_service_port)
                            _logger.debug("PUT-CATEGORY: {} - JDOC: {}".format(y['key'], jdoc))
                            if not isinstance(jdoc, dict):
                                roll_back()
                                raise ValueError('Unable to update Category for name: {} '
                                                 'with requested data: {}'.format(y['key'], data))
                            else:
                                # ROLLBACK: entity, verb, uri, cat name, old data
                                old_data = {}
                                for od in data:
                                    _value = [cfg_item[od] for cfg_item in fl_cat_list if od in cfg_item]
                                    old_data[od] = _value[0]
                                r_tuple = ("CATEGORY", "PUT", "/foglamp/category", x['key'], old_data)
                                roll_back_items.append(r_tuple)
        # Case of New category
        cached_cat_list = [x['key'] for x in cached_cat]
        fl_cat_list = [y['key'] for y in fl_cat]
        list_diff = list(set(cached_cat_list) - set(fl_cat_list))
        _logger.debug("DIFF of New Category list: {}".format(list_diff))
        for ld in list_diff:
            # Create dict for to accept valid payload format on FL API to create new category
            cat_dict = [x for x in cached_cat if ld == x['key']]
            cat_value = {}
            for k, v in json.loads(cat_dict[0]['value']).items():
                # Copying 'value' to 'default' as POST category to FL for a config item 'value' KV pair is not needed
                if 'default' in v:
                    v['default'] = v['value']
                # Ignoring 'value' KV pair from config item
                if 'value' in v:
                    del v['value']
                cat_value[k] = v
            temp_dict = {"key": cat_dict[0]['key'],
                         "description": cat_dict[0]['description'],
                         "displayName": cat_dict[0]['displayName'],
                         "value": cat_value
                         }
            data = temp_dict
            jdoc = Server.make_http_request('POST', '/foglamp/category', payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-CATEGORY: {}".format(jdoc))
            if not isinstance(jdoc, dict):
                roll_back()
                raise ValueError('Unable to create Category with requested data: {}'.format(data))
            else:
                name = jdoc['key']
                # entity, verb, uri, cat_name, empty data
                r_tuple = ("CATEGORY", "DELETE", "/foglamp/category", name, {})
                roll_back_items.append(r_tuple)
        # Case of New Category Children
        for d in list_diff:
            cat_info = [x for x in cached_cat if d == x['key']]
            # Fetch cached children and if found then create children on FL side
            row = db.select_table(db_conn, "SELECT * FROM category_children where child='{}'".format(
                cat_info[0]['key']))
            if row:
                data = {'children': [row[0][1]]}
                jdoc = Server.make_http_request('POST', '/foglamp/category/{}/children'.format(quote(row[0][0])),
                                                payload=data, port=Server._core_service_port)
                _logger.debug("JDOC - POST-CATEGORY CHILDREN: {}".format(jdoc))
                if not isinstance(jdoc, dict):
                    roll_back()
                    raise ValueError('Unable to create Children for category: {} with payload: {}'.format(
                        cat_info[0]['key'], data))

    def add_and_update_schedule():
        # Case of Update Schedule
        for x in cached_sch:
            for y in fl_sch:
                if x['name'] == y['name']:
                    # For matched schedule; Find diff of cached & FL schedule and if found then update on FL side
                    diff = get_diff(y, x)
                    if diff:
                        data = {}
                        for d in diff:
                            data[d] = Schedule.Type[x[d]].value if 'type' in d else x[d]
                            # SPECIAL CASE: if processName is changed then we need to convert into process_name key as FL API accepts that in payload
                            # then we need to disable and re-enabling that instance to reflect the changes only if schedule is enabled
                            if d == 'processName':
                                data['process_name'] = data[d]
                        jdoc = Server.make_http_request('PUT', '/foglamp/schedule/{}'.format(y['id']), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - PUT-SCHEDULE: {}".format(jdoc))
                        if not isinstance(jdoc, dict):
                            roll_back()
                            raise ValueError('Unable to update Schedule for id: {} with requested data: {}'.format(
                                y['id'], data))
                        else:
                            if 'processName' in diff:
                                if y['enabled'] is True:
                                    sch_disable_jdoc = Server.make_http_request(
                                        'PUT', '/foglamp/schedule/{}/disable'.format(y['id']), payload=None,
                                        port=Server._core_service_port)
                                    _logger.debug("JDOC - PUT-DISABLE SCHEDULE: {}".format(sch_disable_jdoc))
                                    # Extra sleep required
                                    time.sleep(WAIT_TIME - 5)
                                    sch_enable_jdoc = Server.make_http_request(
                                        'PUT', '/foglamp/schedule/{}/enable'.format(y['id']), payload=None,
                                        port=Server._core_service_port)
                                    _logger.debug("JDOC - PUT-ENABLE SCHEDULE: {}".format(sch_enable_jdoc))
                                # Remove the process_name key as rollback should match as per FogLAMP Manage API keys
                                del data['process_name']
                            old_data = {}
                            for od in data:
                                if 'type' in od:
                                    _value = Schedule.Type[y[od]].value
                                elif 'repeat' in od:
                                    _value = int(y[od])
                                else:
                                    _value = y[od]
                                old_data[od] = _value
                            r_tuple = ("SCHEDULE", "PUT", "/foglamp/schedule", y['id'], old_data)
                            roll_back_items.append(r_tuple)
        # Case of New Schedule
        cached_sch_list = [x['name'] for x in cached_sch]
        fl_sch_list = [y['name'] for y in fl_sch]
        diff_list = list(set(cached_sch_list) - set(fl_sch_list))
        _logger.debug("DIFF of New Schedule list: {}".format(diff_list))
        for ld in diff_list:
            sch_dict = [sch for sch in cached_sch if ld == sch['name']]
            # Create dict for to accept valid payload format on FL API to create new schedule
            temp = sch_dict[0]
            temp["type"] = Schedule.Type[temp["type"]].value
            temp["day"] = 0 if temp["day"] is None else temp["day"]
            temp["process_name"] = temp["processName"]
            del temp["processName"]
            data = temp
            jdoc = Server.make_http_request('POST', '/foglamp/schedule', payload=data,
                                            port=Server._core_service_port)
            _logger.debug("JDOC - POST-SCHEDULE: {}".format(jdoc))
            if not isinstance(jdoc, dict):
                roll_back()
                raise ValueError('Unable to create Schedule with requested data: {}'.format(data))
            else:
                sch_id = jdoc['schedule']['id']
                # entity, verb, uri, sch_id, empty data
                r_tuple = ("SCHEDULE", "DELETE", "/foglamp/schedule", sch_id, {})
                roll_back_items.append(r_tuple)

    def add_and_update_filters() -> None:
        script_dir = _FOGLAMP_DATA + '/scripts/' if _FOGLAMP_DATA else _FOGLAMP_ROOT + "/data/scripts/"
        for c in get_cached_filters:
            svc_filter_cat_name = ""
            try:
                instance_name_from_filter_cat = next(item for item in get_cached_pipelines
                                                     if c[0] in json.loads(item[1]))
                svc_filter_cat_name = "{}_{}".format(instance_name_from_filter_cat[0], c[0])
            except Exception:
                _logger.error("{} not found in pipelines".format(c[0]))
            _logger.debug("Filter name: {}".format(svc_filter_cat_name))
            if svc_filter_cat_name:
                new_dict = json.loads(c[1])['filter_config']
                new_dict["plugin"] = json.loads(c[1])['plugin']
                _logger.debug("New dict: {}".format(new_dict))
                old_dict = {}
                type_dict = {}
                try:
                    res = next(item for item in fl_cat if item["key"] == svc_filter_cat_name)
                    _logger.debug("Filter Category found: {}".format(res))
                    for k, v in res['value'].items():
                        if k in new_dict:
                            old_dict[k] = v['value']
                            type_dict[k] = v['type']
                    _logger.debug("OLD Filter config: {}".format(old_dict))
                    _logger.debug("New Filter config: {}".format(new_dict))
                    _logger.debug("Type: {}".format(type_dict))
                    diff = get_diff(old_dict, new_dict)
                    _logger.debug("DIFF: {}".format(diff))
                    # SPECIAL CASE
                    # if plugin key is found in filter objet & different then disable schedule with the same category name
                    # Delete the both main filter & standalone categories
                    # recreate the filter category with new plugin config
                    # & then re-enabled the schedule for that instance
                    if 'plugin' in diff:
                        _logger.info("{} plugin change detection case for {} filter category".format(new_dict["plugin"],
                                                                                                     svc_filter_cat_name))
                        instance_schedule_payload = {"schedule_name": instance_name_from_filter_cat[0]}
                        _logger.debug("Disable instance: {} with payload: {}".format(
                            instance_name_from_filter_cat[0], instance_schedule_payload))
                        # Disable instance
                        disable_sch_jdoc = Server.make_http_request(
                            'PUT', '/foglamp/schedule/disable', payload=instance_schedule_payload,
                            port=Server._core_service_port)
                        _logger.debug("JDOC - PUT Disable schedule {}".format(disable_sch_jdoc))
                        # Extra wait time sleep is required to get completely shutdown the instance
                        time.sleep(WAIT_TIME)
                        if 'scheduleId' in disable_sch_jdoc:
                            _logger.debug("DELETE main filter category: {}".format(svc_filter_cat_name))
                            # Delete main filter category
                            delete_filter_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(
                                quote(svc_filter_cat_name)), payload=None, port=Server._core_service_port)
                            _logger.debug("JDOC - DELETE main Filter category: {}".format(delete_filter_cat_jdoc))
                            if 'result' in delete_filter_cat_jdoc:
                                _logger.debug("DELETE standalone filter category: {}".format(c[0]))
                                # Delete standalone category
                                delete_standalone_cat_jdoc = Server.make_http_request(
                                    'DELETE', '/foglamp/category/{}'.format(quote(c[0])), payload=None,
                                    port=Server._core_service_port)
                                _logger.debug("JDOC - DELETE standalone Filter category: {}".format(
                                    delete_standalone_cat_jdoc))
                                if 'result' in delete_standalone_cat_jdoc:
                                    # Recreate filter
                                    filter_config_dict = new_dict.copy()
                                    del filter_config_dict['plugin']
                                    recreate_filter_payload = {"name": c[0], "plugin": new_dict["plugin"],
                                                               "filter_config": filter_config_dict}
                                    _logger.debug("Recreate main {} filter category with payload: {}".format(
                                        svc_filter_cat_name, recreate_filter_payload))
                                    post_recreate_filter_jdoc = Server.make_http_request(
                                        'POST', '/foglamp/filter', payload=recreate_filter_payload,
                                        port=Server._core_service_port)
                                    _logger.debug("JDOC - POST-Recreate Filter: {}".format(post_recreate_filter_jdoc))
                                    if 'filter' in post_recreate_filter_jdoc:
                                        # Below check is to make sure filter recreated with same name which we deleted above
                                        if post_recreate_filter_jdoc['filter'] == c[0]:
                                            # Re-enable the instance
                                            enable_sch_jdoc = Server.make_http_request(
                                                'PUT', '/foglamp/schedule/enable', payload=instance_schedule_payload,
                                                port=Server._core_service_port)
                                            _logger.debug("JDOC - PUT enable schedule {}".format(enable_sch_jdoc))
                                            if 'scheduleId' in enable_sch_jdoc:
                                                _logger.info("Successfully recreated {} filter with {} instance".format(
                                                    svc_filter_cat_name, instance_name_from_filter_cat[0]))
                                            else:
                                                _logger.error("{} schedule enable failed".format(
                                                    instance_name_from_filter_cat[0]))
                                        else:
                                            _logger.error("{} filter is not matched on recreation".format(c[0]))
                                    else:
                                        _logger.error("{} filter recreation failed".format(svc_filter_cat_name))
                                else:
                                    _logger.error("{} category delete failed".format(instance_name_from_filter_cat[0]))
                            else:
                                _logger.error("{} category delete failed".format(svc_filter_cat_name))
                        else:
                            _logger.error("{} schedule disable failed".format(instance_name_from_filter_cat[0]))
                    else:
                        payload = {}
                        for d in diff:
                            payload.update({d: new_dict[d]})
                            if type_dict[d] == 'script':
                                # find script filename & write script content into
                                script_value = new_dict[d]
                                pattern = '__filename__ = "(.*)"'
                                result = re.findall(pattern, script_value)
                                if result:
                                    _logger.debug("Filename found: {}-{}-{}".format(result[0].strip(), len(result[0]),
                                                                                    len(result[0].strip())))
                                    with open("{}{}_script_{}".format(script_dir, svc_filter_cat_name.lower(),
                                                                      result[0].strip()), "w") as script_file:
                                        # Write content to a file
                                        script_file.write(script_value)
                                else:
                                    _logger.error("For category: {} with {} config item unable to detect "
                                                  "__filename__".format(svc_filter_cat_name, d))
                        if payload:
                            bulk_update_jdoc = Server.make_http_request('PUT', '/foglamp/category/{}'.format(
                                quote(svc_filter_cat_name)), payload=payload, port=Server._core_service_port)
                            _logger.debug("JDOC - PUT-CATEGORY: {} with payload: {}".format(bulk_update_jdoc, payload))
                except Exception:
                    _logger.debug("Case of POST Filter for: {} with payload {}".format(c[0], c[1]))
                    """ Special Case:
                        GIVEN: if we create filters at FogLAMP instances before the first time to get registered/managed 
                        at FogLAMP Manage side 
                        WHEN: standalone filter category is being created before the POST filter. 
                              Hence POST request is failed which results in disappearing of filters.
                        THEN: Check if standalone category exists then only call Delete category before new filter 
                              creation.
                    """
                    get_cat_jdoc = Server.make_http_request('GET', '/foglamp/category', payload=None,
                                                            port=Server._core_service_port)
                    synch_fl_categories = [cat['key'] for cat in get_cat_jdoc['categories']]
                    if c[0] in synch_fl_categories:
                        del_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(quote(c[0])),
                                                                payload=None, port=Server._core_service_port)
                        _logger.debug("DELETE-STANDALONE-Filter Category If exists - JDOC: {}".format(del_cat_jdoc))
                    post_filter_jdoc = Server.make_http_request('POST', '/foglamp/filter', payload=json.loads(c[1]),
                                                                port=Server._core_service_port)
                    _logger.debug("POST-Filter - JDOC: {}".format(post_filter_jdoc))
                    # Script case
                    filter_config = json.loads(c[1])['filter_config']
                    _logger.debug("Filter config: {}".format(filter_config))
                    if 'script' in filter_config:
                        """ Find script filename
                            We have predefined contract for the script and it should have following
                            a) filename should be saved & same as def
                            b) __filename__ must be given in the script

                            As per above contract, below is the example:
                            a) filename is "logging.py" and the fucntion name is "def logging(readings):"
                            b) __filename__ = "logging.py"
                            c) Content of logging.py
                            ""
                            Logging of readings in syslogs
                            ""

                            __author__ = "Ashish Jabble"
                            __copyright__ = "Copyright (c) 2021 Dianomic Systems Inc."
                            __license__ = "Apache 2.0"
                            __version__ = "${VERSION}"
                            __filename__ = "logging.py"

                            import logging
                            from logging.handlers import SysLogHandler

                            logger = logging.getLogger(__name__)
                            handler = SysLogHandler(address='/dev/log')
                            formatter = logging.Formatter(fmt='FogLAMP[%(process)d] %(levelname)s: %(name)s: %(message)s')
                            handler.setFormatter(formatter)
                            logger.setLevel(logging.INFO)
                            logger.propagate = False
                            logger.addHandler(handler)


                            ""
                            Method for logging of readings
                            Input is the list of readings
                            ""


                            def logging(readings):
                                global logger
                                logger.info("Readings: {}".format(readings))

                        """
                        _logger.debug(
                            "Filter cat name: {} and script dir path {}".format(svc_filter_cat_name, script_dir))
                        # find script filename & write script content into
                        script_value = filter_config['script']
                        pattern = '__filename__ = "(.*)"'
                        result = re.findall(pattern, script_value)
                        if result:
                            _logger.debug("Filename found: {}-{}-{}".format(result[0].strip(), len(result[0]),
                                                                            len(result[0].strip())))
                            with open("{}{}_script_{}".format(script_dir, svc_filter_cat_name.lower(),
                                                              result[0].strip()), "w") as script_file:
                                # Write content to a file
                                script_file.write(script_value)
                        else:
                            _logger.error("For category: {} unable to detect __filename__".format(
                                svc_filter_cat_name))

    def add_and_update_pipelines() -> None:
        for _pipeline in get_cached_pipelines:
            get_pipeline_jdoc = Server.make_http_request('GET', '/foglamp/filter/{}/pipeline'.format(
                quote(_pipeline[0])), payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - GET Pipeline for {} service: {}".format(_pipeline[0], get_pipeline_jdoc))
            new_pipeline = {"pipeline": json.loads(_pipeline[1])}
            if isinstance(get_pipeline_jdoc, dict):
                old_pipeline = get_pipeline_jdoc['result']['pipeline']
                pipeline_diff = list(set(old_pipeline) - set(new_pipeline['pipeline']))
                uri = "/foglamp/filter/{}/pipeline?allow_duplicates=true&append_filter=false".format(
                    quote(_pipeline[0]))
                put_pipeline_jdoc = Server.make_http_request('PUT', uri, payload=new_pipeline,
                                                             port=Server._core_service_port)
                _logger.debug("JDOC - PUT Pipeline for {} service: {}".format(_pipeline[0], put_pipeline_jdoc))
                for d in pipeline_diff:
                    delete_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(quote(d)),
                                                               payload=None, port=Server._core_service_port)
                    _logger.debug("JDOC - Filter Standalone DELETE CATEGORY: {}".format(delete_cat_jdoc))
            else:
                uri = "/foglamp/filter/{}/pipeline?allow_duplicates=true&append_filter=false".format(
                    quote(_pipeline[0]))
                put_pipeline_jdoc = Server.make_http_request('PUT', uri, payload=new_pipeline,
                                                             port=Server._core_service_port)
                _logger.debug("JDOC - PUT Pipeline for {} service: {}".format(_pipeline[0], put_pipeline_jdoc))

    def add_and_update_certificate_stores() -> None:
        cert_dir_path = "{}{}".format(_FOGLAMP_DATA, "/etc/certs/") if _FOGLAMP_DATA else "{}{}".format(
            _FOGLAMP_ROOT, "/data/etc/certs/")
        # NOTE:
        # We are directly saving files under default cert directory path instead of calling POST cert store via FL-API
        for cc in cached_certs:
            if cc[1]:
                cached_cert_store_for_deletion.append(cc[1])
                if cc[1].endswith('.pem'):
                    _logger.debug("CERT should save under: {}pem/{}".format(cert_dir_path, cc[1]))
                    with open("{}pem/{}".format(cert_dir_path, cc[1]), "w") as pem_cert_file:
                        # Write content to a file
                        pem_cert_file.write(cc[0])
                elif cc[1].endswith('.json'):
                    _logger.debug("CERT should save under: {}json/{}".format(cert_dir_path, cc[1]))
                    with open("{}json/{}".format(cert_dir_path, cc[1]), "w") as json_cert_file:
                        # Write content to a file
                        json_cert_file.write(cc[0])
                else:
                    _logger.debug("CERT should save under: {}{}".format(cert_dir_path, cc[1]))
                    with open("{}{}".format(cert_dir_path, cc[1]), "w") as cert_file:
                        # Write content to a file
                        cert_file.write(cc[0])
            if cc[3]:
                cached_cert_store_for_deletion.append(cc[3])
                if cc[3].endswith('.pem'):
                    _logger.debug("KEY should save under: {}pem/{}".format(cert_dir_path, cc[3]))
                    with open("{}pem/{}".format(cert_dir_path, cc[3]), "w") as pem_key_file:
                        # Write content to a file
                        pem_key_file.write(cc[2])
                else:
                    _logger.debug("KEY should save under: {}{}".format(cert_dir_path, cc[3]))
                    with open("{}{}".format(cert_dir_path, cc[3]), "w") as key_file:
                        # Write content to a file
                        key_file.write(cc[2])
        _logger.debug("Cached Cert store deletion list: {}".format(cached_cert_store_for_deletion))

    def add_and_update_control_script() -> None:
        # Case of Update Control Script
        for x in cached_control_scripts:
            for y in fl_control_scripts:
                if x['name'] == y['name']:
                    payload = {"steps": json.loads(x['steps']), "acl": x['acl']}
                    jdoc = Server.make_http_request('PUT', '/foglamp/control/script/{}'.format(quote(x['name'])),
                                                    payload=payload, port=Server._core_service_port)
                    _logger.debug("For {} with payload, JDOC - PUT-Control script: {}".format(x['name'],
                                                                                              payload, jdoc))
        # Case of New control script
        cached_control_script_list = [ccs['name'] for ccs in cached_control_scripts]
        fl_control_script_list = [y['name'] for y in fl_control_scripts]
        diff_control_script_list = list(set(cached_control_script_list) - set(fl_control_script_list))
        _logger.debug("DIFF of New Control Script list: {}".format(diff_control_script_list))
        for control_script in diff_control_script_list:
            data = [ccs for ccs in cached_control_scripts if control_script == ccs['name']]
            payload = {"name": data[0]['name'], "steps": json.loads(data[0]['steps'])}
            if len(data[0]['acl']):
                payload['acl'] = data[0]['acl']
            jdoc = Server.make_http_request('POST', '/foglamp/control/script', payload=payload,
                                            port=Server._core_service_port)
            _logger.debug("For {} with payload, JDOC - POST-Control script: {}".format(control_script, data, jdoc))

    def add_and_update_control_acl() -> None:
        # Case of Update Control ACL
        for x in cached_control_acls:
            for y in fl_control_acls:
                if x['name'] == y['name']:
                    payload = {"service": json.loads(x['service']), "urls": json.loads(x['url'])}
                    jdoc = Server.make_http_request('PUT', '/foglamp/ACL/{}'.format(quote(x['name'])),
                                                    payload=payload, port=Server._core_service_port)
                    _logger.debug("For {} with payload, JDOC - PUT-Control ACL: {}".format(x['name'], payload, jdoc))
        # Case of New control ACL
        cached_control_acl_list = [ccs['name'] for ccs in cached_control_acls]
        fl_control_acl_list = [y['name'] for y in fl_control_acls]
        diff_control_acl_list = list(set(cached_control_acl_list) - set(fl_control_acl_list))
        _logger.debug("DIFF of New Control ACL list: {}".format(diff_control_acl_list))
        for control_acl in diff_control_acl_list:
            data = [cca for cca in cached_control_acls if control_acl == cca['name']]
            payload = {"name": data[0]['name'], "service": json.loads(data[0]['service']),
                       "url": json.loads(data[0]['url'])}
            jdoc = Server.make_http_request('POST', '/foglamp/ACL', payload=payload, port=Server._core_service_port)
            _logger.debug("For {} with payload, JDOC - POST-Control ACL: {}".format(control_acl, data, jdoc))

    def update_attach_acls() -> None:
        get_attached_acls = db.select_table(db_conn, "SELECT * FROM attach_acl")
        schedules = [s['name'] for s in cached_sch if s['type'] in ('STARTUP', 1)]
        for gaa in get_attached_acls:
            if gaa[0] in schedules:
                if gaa[1]:
                    data = {"acl_name": gaa[1]}
                    jdoc = Server.make_http_request('PUT', '/foglamp/service/{}/ACL'.format(quote(gaa[0])),
                                                    payload=data, port=Server._core_service_port)
                    _logger.debug("ATTACH_SERVICE WITH ACL: {} - JDOC: {}".format(gaa[0], jdoc))
                else:
                    data = {"acl_name": ""}
                    jdoc = Server.make_http_request('DELETE', '/foglamp/service/{}/ACL'.format(quote(gaa[0])),
                                                    payload=data, port=Server._core_service_port)
                    _logger.debug("DETACH_SERVICE WITH ACL: {} - JDOC: {}".format(gaa[0], jdoc))

    def add_and_update_control_pipeline() -> None:
        # Case of Update Control Pipeline
        for x in cached_cp:
            for y in fl_control_pipelines:
                if x['name'] == y['name']:
                    pipeline_filters = x['filters']
                    conv_pipeline_filters = []
                    if pipeline_filters:
                        val = eval(pipeline_filters.strip("]["))
                        conv_pipeline_filters = list(val) if isinstance(val, tuple) else val.split()
                        for f in get_cached_filters:
                            if f[0] in pipeline_filters:
                                post_filter_jdoc = Server.make_http_request(
                                    'POST', '/foglamp/filter', payload=json.loads(f[1]), port=Server._core_service_port)
                                _logger.debug("PUT-Filter Control Pipeline - JDOC: {}".format(post_filter_jdoc))
                    payload = {
                        "name": x['name'],
                        "source": {"type": x['stype'], "name": x['sname']},
                        "destination": {"type": x['dtype'], "name": x['dname']},
                        "enabled": True if x['enabled'] else False,
                        "execution": x['execution'],
                        "filters": conv_pipeline_filters
                    }
                    jdoc = Server.make_http_request('PUT', '/foglamp/control/pipeline/{}'.format(y['id']),
                                                    payload=payload, port=Server._core_service_port)
                    _logger.debug("For {} with payload, JDOC - PUT-Control Pipeline: {}".format(x['name'],
                                                                                                payload, jdoc))
        # Case of New control Pipeline
        cached_control_pipeline_list = [cp['name'] for cp in cached_cp]
        fl_control_pipeline_list = [y['name'] for y in fl_control_pipelines]
        diff_control_pipeline_list = list(set(cached_control_pipeline_list) - set(fl_control_pipeline_list))
        _logger.debug("DIFF of New Control Pipeline list: {}".format(diff_control_pipeline_list))
        for control_pipeline in diff_control_pipeline_list:
            data = [ccp for ccp in cached_cp if control_pipeline == ccp['name']]
            pipeline_filters = data[0]['filters']
            conv_pipeline_filters = []
            if pipeline_filters:
                val = eval(pipeline_filters.strip("]["))
                conv_pipeline_filters = list(val) if isinstance(val, tuple) else val.split()
                for f in get_cached_filters:
                    if f[0] in pipeline_filters:
                        post_filter_jdoc = Server.make_http_request('POST', '/foglamp/filter', payload=json.loads(f[1]),
                                                                    port=Server._core_service_port)
                        _logger.debug("POST-Filter In case of Control Pipeline - JDOC: {}".format(post_filter_jdoc))

            payload = {
                "name": data[0]['name'],
                "source": {"type": data[0]['stype'], "name": data[0]['sname']},
                "destination": {"type": data[0]['dtype'], "name": data[0]['dname']},
                "enabled": True if data[0]['enabled'] else False,
                "execution": data[0]['execution'],
                "filters": conv_pipeline_filters
            }
            jdoc = Server.make_http_request('POST', '/foglamp/control/pipeline', payload=payload,
                                            port=Server._core_service_port)
            _logger.debug("For {} with payload: {}, JDOC - POST-Control Pipeline: {}".format(
                control_pipeline, payload, jdoc))

    def delete_entries_from_fl() -> None:
        fl_cat_list = [y1['key'] for y1 in fl_cat]
        cached_cat_list = [x1['key'] for x1 in cached_cat]
        diff_list_cat = list(set(fl_cat_list) - set(cached_cat_list))
        delete_cat_list = []
        for delete_cat in diff_list_cat:
            # Ignore run time categories while deleting service to save attempt
            if not (delete_cat.endswith(("Advanced", "Security", " Filters"))):
                try:
                    sch_id = next(item['id'] for item in fl_sch if
                                  item["name"] == delete_cat and item["processName"] in ['south_c', 'north_C'])
                    if sch_id:
                        delete_svc_jdoc = Server.make_http_request('DELETE', '/foglamp/service/{}'.format(
                            quote(delete_cat)), payload=None, port=Server._core_service_port)
                        _logger.debug("JDOC - DELETE SERVICE: {}".format(delete_svc_jdoc))
                        # Extra wait time needed to complete recursively deletion
                        time.sleep(WAIT_TIME - 10)
                except Exception:
                    try:
                        sch_id = next(item['id'] for item in fl_sch if
                                      item["name"] == delete_cat and item["processName"] == 'north_c')
                        if sch_id:
                            delete_task_jdoc = Server.make_http_request('DELETE', '/foglamp/scheduled/task/{}'.format(
                                quote(delete_cat)), payload=None, port=Server._core_service_port)
                            _logger.debug("JDOC - DELETE TASK: {}".format(delete_task_jdoc))
                            # Extra wait time needed to complete recursively deletion
                            time.sleep(WAIT_TIME - 10)
                    except Exception:
                        # Don't delete Parent categories Notification, North, South
                        if delete_cat not in ['Notifications', 'North', 'South']:
                            delete_cat_list.append(delete_cat)
        # Filters
        get_fl_filters = Server.make_http_request('GET', '/foglamp/filter', payload=None,
                                                  port=Server._core_service_port)
        fl_filter = get_fl_filters["filters"]
        fl_filter_list = [y1['name'] for y1 in fl_filter]
        temp = fl_filter_list.copy()
        for f in temp:
            get_fl_filter_user = Server.make_http_request('GET', '/foglamp/filter/{}'.format(quote(f)), payload=None,
                                                          port=Server._core_service_port)
            if isinstance(get_fl_filter_user, dict):
                fl_filter_user = get_fl_filter_user["filter"]["users"]
                for fuser in fl_filter_user:
                    fl_filter_list.append('{}_{}'.format(fuser, f))
                    fl_filter_list.append('{} Filters'.format(fuser))
        _logger.debug("DELETE Filter Name List: {}".format(fl_filter_list))
        _logger.debug("DELETE CAT LIST: {}".format(delete_cat_list))
        others_delete_list = list(set(delete_cat_list) - set(fl_filter_list))
        _logger.debug("DELETE OTHERS LIST: {}".format(others_delete_list))
        for c in others_delete_list:
            delete_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(quote(c)),
                                                       payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE CATEGORY: {}".format(delete_cat_jdoc))
        # Schedules
        cached_sch_list = [x2['name'] for x2 in cached_sch]
        fl_sch_list = [y2['name'] for y2 in fl_sch]
        diff_list_sch = list(set(fl_sch_list) - set(cached_sch_list))
        _logger.debug("DIFF of Delete Schedule list: {}".format(diff_list_sch))
        others_delete_sch_list = list(set(diff_list_sch) - set(diff_list_cat))
        _logger.debug("Others DELETE Schedule LIST: {}".format(others_delete_sch_list))
        for delete_cat in others_delete_sch_list:
            try:
                sch_id = next(item['id'] for item in fl_sch if item["name"] == delete_cat)
                delete_sch_jdoc = Server.make_http_request('DELETE', '/foglamp/schedule/{}'.format(sch_id),
                                                           payload=None, port=Server._core_service_port)
                _logger.debug("JDOC - DELETE SCHEDULE: {}".format(delete_sch_jdoc))
            except Exception:
                pass
        # Cert Stores
        diff_cert_stores_list = list(set(fl_certs_without_inbuilt) - set(cached_cert_store_for_deletion))
        _logger.debug("DELETE cert stores list: {}".format(diff_cert_stores_list))
        for cert_name in diff_cert_stores_list:
            delete_cert_store_jdoc = Server.make_http_request('DELETE', '/foglamp/certificate/{}'.format(cert_name),
                                                              payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE Cert store: {}".format(delete_cert_store_jdoc))
        # control script
        cached_control_script_list = [ccs['name'] for ccs in cached_control_scripts]
        fl_control_script_list = [y['name'] for y in fl_control_scripts]
        diff_control_script_list = list(set(fl_control_script_list) - set(cached_control_script_list))
        _logger.debug("DIFF of Delete Control Script list: {}".format(diff_control_script_list))
        for delete_control_script in diff_control_script_list:
            delete_control_script_jdoc = Server.make_http_request('DELETE', '/foglamp/control/script/{}'.format(
                quote(delete_control_script)), payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE CONTROL SCRIPT: {}".format(delete_control_script_jdoc))
        # control ACL
        cached_control_acl_list = [ccs['name'] for ccs in cached_control_acls]
        fl_control_acl_list = [y['name'] for y in fl_control_acls]
        diff_control_acl_list = list(set(fl_control_acl_list) - set(cached_control_acl_list))
        _logger.debug("DIFF of Delete Control ACL list: {}".format(diff_control_acl_list))
        for delete_control_acl in diff_control_acl_list:
            delete_control_acl_jdoc = Server.make_http_request('DELETE', '/foglamp/ACL/{}'.format(
                quote(delete_control_acl)), payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE CONTROL ACL: {}".format(delete_control_acl_jdoc))

        # control Pipeline
        # TODO: FOGL-8100
        cached_control_pipeline_list = [ccs['name'] for ccs in cached_cp]
        fl_control_pipeline_list = [y['name'] for y in fl_control_pipelines]
        diff_control_pipeline_list = list(set(fl_control_pipeline_list) - set(cached_control_pipeline_list))
        _logger.debug("DIFF of Delete Control Pipeline list: {}".format(diff_control_pipeline_list))
        for delete_control_pipeline in diff_control_pipeline_list:
            cpid = [y['id'] for y in fl_control_pipelines if y['name'] == delete_control_pipeline]
            delete_control_acl_jdoc = Server.make_http_request('DELETE', '/foglamp/control/pipeline/{}'.format(cpid[0]),
                                                               payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE CONTROL Pipeline: {}".format(delete_control_acl_jdoc))

    def roll_back() -> None:
        _logger.debug("ROLL BACK ITEMS: {}".format(roll_back_items))
        for item in roll_back_items:
            _logger.debug("ROLL BACK item: {}".format(item))
            entity = item[0]
            verb = item[1]
            uri = item[2]
            if entity == 'CATEGORY':
                cat_name = quote(item[3])
                if verb == 'PUT':
                    data = {}
                    for k, v in item[4].items():
                        data[k] = v
                    if data:
                        jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, cat_name), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - ROLL BACK PUT-CATEGORY: {}".format(jdoc))
                elif verb == 'DELETE':
                    jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, cat_name), payload={},
                                                    port=Server._core_service_port)
                    _logger.debug("JDOC - ROLL BACK DELETE-CATEGORY: {}".format(jdoc))
                else:
                    _logger.error("For entity {} invalid verb found: {} when rollback".format(entity, verb))
            elif entity == 'SCHEDULE':
                sch_id = item[3]
                if verb == 'PUT':
                    data = {}
                    for k, v in item[4].items():
                        data[k] = v
                    if data:
                        jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, sch_id), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - ROLL BACK PUT-SCHEDULE: {}".format(jdoc))
                elif verb == 'DELETE':
                    jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, sch_id), payload={},
                                                    port=Server._core_service_port)
                    _logger.debug("JDOC - ROLL BACK DELETE-SCHEDULE: {}".format(jdoc))
                else:
                    _logger.error("For entity {} invalid verb found: {} when rollback".format(entity, verb))
            else:
                _logger.error("For Rollback, bad entity found: {}".format(entity))

    def audit_marker_entry() -> None:
        time.sleep(1)
        payload = {
            "source": "AUMRK",
            "severity": "INFORMATION",
            "details": {"action": "deployment", "entity": "Poll Agent",
                        "message": "Successfully deployed instance '{}' to version '{}'".format(
                            fl_svc_name, fl_current_version),
                        "status": "success", "version": fl_current_version}
        }
        jdoc = Server.make_http_request('POST', '/foglamp/audit', payload=payload, port=Server._core_service_port)
        _logger.debug("Audit Marker Payload: {}, JDOC - POST AUDIT ENTRY: {}".format(payload, jdoc))

    install_software()
    install_python_packages()
    update_buckets()
    add_and_update_category()
    add_and_update_schedule()
    add_and_update_filters()
    add_and_update_pipelines()
    add_and_update_certificate_stores()
    add_and_update_control_acl()
    add_and_update_control_script()
    add_and_update_control_pipeline()
    delete_entries_from_fl()
    update_attach_acls()
    audit_marker_entry()

    # Set is_synched to True for instance_version table
    query = "UPDATE instance_version SET is_synched = 't'"
    db.update_table(db_conn, query)


def get_fogman_oauth2_access_token(fogman_host: str, username: str, password: str):
    def get_oauth2_services():
        oauth2_svc_jdoc = _make_https_request(fogman_host, 'GET', '/oauth2/services', payload={})
        if oauth2_svc_jdoc[0] == 200:
            return oauth2_svc_jdoc[1]['oauth2_services'][0]
        else:
            _logger.error("GET /oauth2/services request failed. Got {} HTTP status code and reason {}".format(
                oauth2_svc_jdoc[0], oauth2_svc_jdoc[1]))
            raise

    def get_oauth2_authorize():
        _type = 'token'
        _scope = 'all'
        url = "{}?client_id={}&redirect_uri=http%3A%2F%2F{}%3A12345&response_type={}&scope={}&state={}".format(
            authorize_uri, client_id, fogman_host, _type, _scope, base64_oauth2_authorize_state)
        conn = http.client.HTTPSConnection(fogman_host, server_port, context=ssl._create_unverified_context())
        conn.request('GET', url, body=None)
        authorize_response = conn.getresponse()
        if authorize_response.status == 302:
            # Extract header fields info which we need during oAuth2 login in next step
            _location = authorize_response.getheader('Location')
            _cookie = authorize_response.getheader('Set-Cookie').split(';')[0]
            return _location, _cookie
        else:
            _logger.error("GET /oauth2/authorize request failed. Got {} HTTP status code and reason {}".format(
                authorize_response.status, authorize_response.reason))
            raise

    def post_login():
        headers = {"Content-Type": "application/x-www-form-urlencoded", "Cookie": "{}".format(cookie)}
        # base64 string for password
        pwd = base64.b64encode(hashlib.sha512(password.encode('utf-8')).digest()).decode("utf-8")
        url = "{}?username={}&password={}".format(location, username, pwd)
        conn = http.client.HTTPSConnection(fogman_host, server_port, context=ssl._create_unverified_context())
        conn.request('POST', url, body=None, headers=headers)
        login_response = conn.getresponse()
        if login_response.status == 302:
            # Access token in the URI fragment after #char
            loc = login_response.getheader('Location').split("#")
            # State in the URI fragment And verify with which we sent during GET /oauth2/authorize
            state = loc[1].split('&state=')[1].split('&')[0]
            if state == base64_oauth2_authorize_state:
                return loc[1].split('&')[0].replace('access_token=', '')
            else:
                _logger.error("State parameter does not match that one sent in /oauth2/authorize request.")
                raise
        else:
            _logger.error("POST /oauth2/login request failed. Got {} HTTP status code and reason {}".format(
                login_response.status, login_response.reason))
            raise

    # Encoded base64 string for state (i.e MGT-LOGIN_uuid)
    base64_oauth2_authorize_state = base64.b64encode("MGT-LOGIN_{}".format(
        uuid.uuid4()).encode('ascii')).decode('ascii')
    # GET oAuth2 services from FogLAMP Manage API
    fogman_svc = get_oauth2_services()
    # Below are the values are required to be passed for oAuth2 authorize request
    client_id = fogman_svc['oauth2_app_id']
    server_url = fogman_svc['oauth2_server_name']
    server_port = server_url.split(':')[-1]
    authorize_uri = fogman_svc['oauth2_authorize_uri']
    # oAuth2 Authorize Request
    location, cookie = get_oauth2_authorize()
    # oAuth2 Login Request
    oauth2_token = post_login()
    return oauth2_token


def upload_bucket_along_with_association(bucket_url: str, payload: dict, fm_host: str, version: str, fl_svc_name: str,
                                         req_header: dict) -> None:
    import requests
    from requests_toolbelt.multipart.encoder import MultipartEncoder


    mp_encoder = MultipartEncoder(
        fields={
            'attributes': json.dumps(payload['attributes']),
            # plain file object, no filename or mime type produces a
            # Content-Disposition header with just the part name
            'bucket': (payload['file'], open(payload['file'], 'rb'), 'text/plain')
        }
    )
    headers = {'Content-Type': mp_encoder.content_type}
    headers.update(req_header)
    r = requests.post(
        bucket_url,
        data=mp_encoder,  # The MultipartEncoder is posted as data, don't use files=...!
        # The MultipartEncoder provides the content-type header with the boundary:
        headers=headers,
        verify=False
    )
    if r.status_code not in range (200, 209):
        _logger.error("Add bucket failed for url: {}".format(bucket_url))
    else:
        _logger.debug("Bucket added successfully for url: {} with attributes: {}".format(
            bucket_url, payload['attributes']))
        response = json.loads(r.text)
        bucket_id = payload['bucket']
        fl_bucket_id = response['bucket']
        if req_header:
            bucket_id = response['bucket']
            fl_bucket_id = payload['bucket']
        associate_bucket_url = "/manage/version/{}/bucket/{}/foglamp/{}".format(version, bucket_id, fl_svc_name)
        associate_bucket_jdoc = _make_https_request(fm_host, 'POST', associate_bucket_url,
                                                    payload={"foglamp_bucket": fl_bucket_id})
        _logger.debug("For URL: {}; JDOC associate bucket: {}".format(
            associate_bucket_url, associate_bucket_jdoc))


def is_service_running(Server, svc_type):
    found_occurrence = False
    max_attempts = 5
    retry_count = 0
    while not found_occurrence and retry_count < max_attempts:
        retry_count += 1
        get_services = Server.make_http_request('GET', '/foglamp/service', payload=None,
                                                port=Server._core_service_port)
        fl_services_from_registry = get_services["services"]
        for svc in fl_services_from_registry:
            if svc["type"] == svc_type and svc["status"] == "running":
                found_occurrence = True
                break
        time.sleep(5)
    return found_occurrence


async def poll(request: web.Request) -> web.Response:
    # To avoid circular import
    from foglamp.services.management.server import Server

    if Server._is_poll_running:
        msg = "FogMan Agent Poll schedule is already running. Hence poll request is rejected."
        _logger.warning(msg)
        return web.json_response({"message": msg})

    if Server._config['enable']['value'] == 'false':
        msg = "operation cannot be executed as service configuration in disabled state."
        _logger.warning(msg)
        return web.json_response({"message": msg})

    fogman_host = Server._config['manager']['value']
    response = {"message": {}}
    try:
        Server._is_poll_running = True
        # Get service name of FogLAMP instance
        svc_category_jdoc = Server.make_http_request('GET', '/foglamp/category/service', payload=None,
                                                     port=Server._core_service_port)
        fl_service_name = svc_category_jdoc['name']['value']
        # NOTE:
        # oAuth2 implementation as per Specs - All clients (GUI, CLI, Mgt) will use same fogman CLIENT/APP ID
        # Some time in the future we support using external auth services then there will be more than one.
        # For the time we are safe to assume just a single entry (i.e /oauth2/services ) in the response.
        # Get oAuth2 Access Token

        Server._access_token = get_fogman_oauth2_access_token(fogman_host=fogman_host,
                                                              username=Server._config['username']['value'],
                                                              password=Server._config['password']['value'])
        _logger.debug("Access_token: {}".format(Server._access_token))
        # NOTE:
        # 1) It is expected to register FL instance with the name in service category.
        #    If already known to FogLAMP Manage, then what name FogLAMP should use (and update service category before add?)
        #  2) knowing by name is enough? what if address mismatch
        #     Let say I had F1 name (and it was you who registred already with F1) ;
        #     checking with name will say oh yeah you are known to me.
        #  Probable solution for #2:
        #   a) Fetch the version /status/foglamp/F1
        #   b) Find the address manage/version/{version}/foglamp
        # Check FL is known to the FogLAMP Manage
        fl_status_jdoc = _make_https_request(fogman_host, 'GET', '/status/foglamp/{}'.format(fl_service_name),
                                             payload={})
        if fl_status_jdoc[0] == 200:
            deployed_version = fl_status_jdoc[1]['status']['deployed']['version']
            response['message'] = _poll_data(fogman_host, fl_service_name)
        elif fl_status_jdoc[0] == 404:
            # FL node is NOT known to FogLAMP Manage
            if Server._push_state in ("STARTED", "COMPLETED"):
                msg = "404: FL data is already pushed to FogLAMP Manage. Hence poll request is rejected."
                _logger.warning(msg)
                return web.json_response({"message": msg})
            is_pushed = _add_fl_and_push_data_to_fogman(fogman_host, fl_service_name)
            response['message'] = "FL data pushed to FogLAMP Manage successfully!" if is_pushed \
                else "Failed to Push data to FogLAMP Manage"
        elif fl_status_jdoc[0] == 400:
            # If no current version found, attached FL instance to default 1.0.0 version and then data push to FogLAMP Manage
            if fl_status_jdoc[1]['error'] == "Current version not found":
                if Server._push_state in ("STARTED", "COMPLETED"):
                    msg = "400: FL data is already pushed to FogLAMP Manage. Hence poll request is rejected."
                    _logger.warning(msg)
                    return web.json_response({"message": msg})
                is_pushed = _add_fl_and_push_data_to_fogman(fogman_host, fl_service_name)
                response['message'] = "FL data pushed to FogLAMP Manage successfully!" if is_pushed \
                    else "Failed to Push data to FogLAMP Manage"
            else:
                raise RuntimeError(fl_status_jdoc[1]['error'])
        else:
            raise RuntimeError(fl_status_jdoc[1]['error'])
    except RuntimeError as err:
        response['message'] = str(err)
        _logger.error("Internal Server error: {}".format(response))
        return web.HTTPInternalServerError(reason=response, body=json.dumps(response))
    except Exception as ex:
        response['message'] = "{}. See detailed exception message in syslog.".format(str(ex))
        _logger.exception("Poll operation failed due to reason: {}".format(str(ex)))
        return web.HTTPInternalServerError(reason=str(ex), body=json.dumps(response))
    else:
        return web.json_response(response)
    finally:
        Server._is_poll_running = False

