import os
import logging
import asyncio
import time
import json
import http.client
import ssl

from aiohttp import web
from foglamp.common import logger
from foglamp.services.management.routes import management, service
from foglamp.common.common import _FOGLAMP_ROOT
from foglamp.services.management import db
from foglamp.services.management.common import utils

__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2020 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_logger = logger.setup(__name__, level=logging.INFO)


class Server:
    _host = '0.0.0.0'
    """ Microservice Host IP """

    _management_port = 0
    """ Microservice management port """

    _service_port = 0
    """ Microservice service port """

    _service_name = "FogLAMP Manage Agent"
    """ Microservice name """

    _core_mgt_host = '0.0.0.0'
    """ Core management host name """

    _core_mgt_port = 0
    """ Core management port """

    _core_service_port = 0
    """ Core service port """

    _service_uuid = None
    """ Microservice uuid """

    _schedule_uuid_for_task = None
    """ Schedule id for manage tasks """

    _is_tls_enabled = False
    """ TLS enable - make secure calls over HTTPS connection when True """

    _auth_token = None
    """ FogLAMP auth token when authentication is mandatory and to be used with Admin API"""

    _start_time = time.time()
    """ Start time of management process """

    _access_token = None
    """ FogLAMP Manager/FogLAMP Manage oAuth2 access token """

    _DEFAULT_CONFIG = {
        'manager': {
            'description': 'Hostname or IP address of the FogLAMP Manage instance managing this FogLAMP instance',
            'type': 'string',
            'default': 'fogman',
            'displayName': 'FogLAMP Manage host',
            'order': '1'
        },
        'enable': {
            'description': 'Enable the management of this node',
            'type': 'boolean',
            'default': 'true',
            'displayName': 'Enable',
            'order': '2'
        },
        'username': {
            'description': 'Username for oAuth2 login',
            'type': 'string',
            'default': 'admin',
            'displayName': 'Username',
            'order': '3'
        },
        'password': {
            'description': 'Password for oAuth2 login',
            'type': 'password',
            'default': 'fogman',
            'displayName': 'Password',
            'order': '4'
        }
    }
    """ Default configuration """
    _config = _DEFAULT_CONFIG
    """ Service configuration """
    _is_poll_running = False
    """ When True the FogMan Agent Poll schedule will be in restricted mode.
    If a previous request is still running then new request will be rejected."""
    _push_state = "UNKNOWN"
    """ Push state when FL node is unknown.
    Possible values are STARTED, COMPLETED, ABORTED, FAILED, UNKNOWN"""

    loop = None
    service_app, service_app_server, service_app_server_handler = None, None, None
    management_app, management_app_server, management_app_server_handler = None, None, None

    @classmethod
    def make_http_request(cls, action, uri, payload, port=None, headers=None):
        port = cls._core_mgt_port if port is None else port
        data = None if payload is None else json.dumps(payload)
        # add comment for condition port != cls._core_mgt_port
        if cls._is_tls_enabled and port != cls._core_mgt_port:
            conn = http.client.HTTPSConnection(cls._core_mgt_host, port, context=ssl._create_unverified_context())
        else:
            conn = http.client.HTTPConnection("{}:{}".format(cls._core_mgt_host, port))

        if cls._auth_token:
            # header key is case insensitive?
            headers = {"authorization": cls._auth_token}
        if headers:
            conn.request(action, uri, body=data, headers=headers)
        else:
            conn.request(action, uri, body=data)

        r = conn.getresponse()
        if r.status == 412:
            pass
        elif r.status in range(400, 500):
            _logger.error("Could not {}: {}, Client error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
        elif r.status in range(500, 600):
            _logger.error("Could not {}: {}, Server error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
        res = r.read().decode()
        # TODO: At the moment with 200 & 412 http status code only; Once FOGL-4547 is resolved will remove that check
        response = json.loads(res) if r.status in (200, 412) else res
        return response

    @classmethod
    def register_service(cls, service_registration_payload, registration_token):
        if registration_token:
            service_registration_payload['token'] = registration_token
        else:
            _logger.error("Single use registration token is missing")
            return
        response = cls.make_http_request('POST', '/foglamp/service', service_registration_payload)
        try:
            cls._service_uuid = response["id"]
        except Exception as ex:
            _logger.exception("Could not register the microservice, From request: {}, Reason: {}".format(json.dumps(
                service_registration_payload), str(ex)))
            raise
        else:
            return response

    @classmethod
    def create_category(cls):
        payload = {"key": cls._service_name, "description": "FogLAMP Manage Agent service",
                   "value": cls._DEFAULT_CONFIG}
        response = cls.make_http_request('POST', '/foglamp/service/category', payload)
        try:
            cls._config = response['value']
        except Exception as ex:
            _logger.exception("Could not create a category, From request: {}, Reason: {}".format(json.dumps(payload),
                                                                                                 str(ex)))
            raise
        else:
            return response

    @classmethod
    def register_interest(cls):
        payload = {"category": cls._service_name, "service": cls._service_uuid}
        response = cls.make_http_request('POST', '/foglamp/interest', payload)
        try:
            response["id"]
        except Exception as ex:
            _logger.exception("Could not create a interest register record, From request: {}, Reason: {}".format(
                json.dumps(payload), str(ex)))
            raise
        else:
            return response

    @classmethod
    def add_schedule_for_task(cls, host_id):
        sch_name = "FogMan Agent Poll"
        process_name = "manage"

        def _check_scheduled_process_entry():
            is_found = False
            jdoc = cls.make_http_request('GET', '/foglamp/schedule/process', payload=None, port=cls._core_service_port)
            for process in jdoc['processes']:
                if process == process_name:
                    is_found = True
            return is_found

        def _check_schedule_entry():
            is_found = False
            jdoc = cls.make_http_request('GET', '/foglamp/schedule', payload=None, port=cls._core_service_port)
            for sch in jdoc['schedules']:
                if sch["name"] == sch_name:
                    cls._schedule_uuid_for_task = sch["id"]
                    is_found = True
            return is_found

        ret_val_for_scheduled_process = _check_scheduled_process_entry()
        if not ret_val_for_scheduled_process:
            # Add scheduled process entry
            payload = {"process_name": process_name, "script": '["tasks/manage"]'}
            response = cls.make_http_request('POST', '/foglamp/schedule/process', payload, port=cls._core_service_port)
        else:
            msg = "Scheduled Process entry already exists for {}".format(process_name)
            _logger.warning(msg)
            response = {"message": msg}
        try:
            if response["message"]:
                ret_val_for_schedule = _check_schedule_entry()
                if not ret_val_for_schedule:
                    # convert_hex_to_decimal - last two bits of hostid
                    _range = int(host_id[-2:], 16)
                    SCHEDULE_TIME_MINS_OFFSET = 180
                    # midnight to 3 AM (i.e. midnight to midnight + SCHEDULE_TIME_MINS_OFFSET # in minutes)
                    sch_time = (_range % SCHEDULE_TIME_MINS_OFFSET) * 60
                    # if sch_time is 0 then set it to 1 And this will allow to add at 00:00:01
                    sch_time = 1 if sch_time == 0 else sch_time
                    repeat = 24 * 60 * 60
                    payload = {"type": 2, "name": sch_name, "process_name": "manage", "repeat": repeat,
                               "time": sch_time, "enabled": "true"}
                    response = cls.make_http_request('POST', '/foglamp/schedule', payload, port=cls._core_service_port)
                    cls._schedule_uuid_for_task = response['schedule']['id']
                else:
                    msg = "Schedule entry already exists for {} and having id {}".format(sch_name,
                                                                                         cls._schedule_uuid_for_task)
                    _logger.warning(msg)
                    response = {"message": msg}
        except (KeyError, Exception) as ex:
            _logger.exception("Could not create a schedule record, From request: {}, Reason: {}".format(
                json.dumps(payload), str(ex)))
            raise
        else:
            return response

    @classmethod
    def start_app(cls, app, host, port, ssl_ctx=None):
        if cls.loop is None:
            cls.loop = asyncio.get_event_loop()
        handler = app.make_handler()
        coro = cls.loop.create_server(handler, host, port, ssl=ssl_ctx)
        server = cls.loop.run_until_complete(coro)
        return server, handler

    @classmethod
    async def stop_app(cls):

        def _unregister_service():
            response = cls.make_http_request(action='DELETE', uri='/foglamp/service/{}'.format(
                cls._service_uuid), payload=None)
            try:
                response["id"]
            except Exception as ex:
                _logger.exception("Could not unregister the microservice, Reason: {}".format(str(ex)))
                raise

        def _disable_and_delete_schedule():
            try:
                response = cls.make_http_request(action='PUT', uri='/foglamp/schedule/{}/disable'.format(
                    cls._schedule_uuid_for_task), payload=None, port=cls._core_service_port)
                # NOTE: When FogLAMP stop/restart/shutdown - Skip disable and delete FogMan Agent Poll Schedule task.
                #       As Scheduler stops and API returns response {'error': {'message': '[NotReadyError] '}}
                if isinstance(response, str):
                    response = json.loads(response)
                if 'error' in response:
                    if 'message' in response['error']:
                        if 'NotReadyError' in response['error']['message']:
                            pass
                else:
                    sch_id = response["scheduleId"]
                    res = cls.make_http_request(action='DELETE', uri='/foglamp/schedule/{}'.format(
                        cls._schedule_uuid_for_task), payload=None, port=cls._core_service_port)
                    sch_id = res["id"]
            except Exception as ex:
                _logger.exception("Could not disable or delete the schedule id {}, Reason: {}".format(
                    cls._schedule_uuid_for_task, str(ex)))
                raise

        _disable_and_delete_schedule()

        cls.service_app_server.close()
        await cls.service_app_server.wait_closed()
        # await cls.service_app.shutdown()
        # await cls.service_app_server_handler.shutdown(5.0)
        await cls.service_app.cleanup()

        cls.management_app_server.close()
        await cls.management_app_server.wait_closed()
        # await cls.management_app.shutdown()
        # await cls.management_app_server_handler.shutdown(5.0)
        await cls.management_app.cleanup()

        _unregister_service()

        cls.loop.stop()

    @classmethod
    def get_version(cls):
        try:
            path = "{}/python/foglamp/services/management/VERSION".format(_FOGLAMP_ROOT)
            my_lines = []
            with open(path, 'rt') as file:
                for f in file:  # For each line, read to a string,
                    my_lines.append(f)
            # Management schema always lies in first line. For example: schema_version=1
            return my_lines[0].replace('\n', '').split('=')[1]
        except (FileNotFoundError, IOError, TypeError):
            _logger.exception("Management VERSION file is NOT found and expected path is {}".format(path))
            raise

    @classmethod
    def run(cls, svc_name, core_mgt_host, core_mgt_port, registration_token=None):
        if registration_token is None:
            # log notice/error and exit
            pass

        cls._service_name = svc_name
        cls._core_mgt_host = core_mgt_host
        cls._core_mgt_port = core_mgt_port
        mgt_schema_version = int(cls.get_version())
        _logger.info("Management Schema Version is: {}".format(mgt_schema_version))
        # Check DB File existence
        if os.path.isfile(db._DB_PATH):
            _logger.info("Management SQLite3 database '{}' ready.".format(db._DB_PATH))
            db_conn = db.create_connection(db._DB_PATH)
            with db_conn:
                row = db.select_table(db_conn, "SELECT id from version")
                db_version = int(row[0][0])
                if mgt_schema_version > db_version:
                    _logger.info("Management DB schema upgrading from [{}] to [{}]...".format(
                        db_version, mgt_schema_version))
                    """ If mgt_schema_version=5; db_version=2 then;
                        run sql files (3.sql, 4.sql, 5.sql) from path data/migration/up
                    """
                    for sql_version in range(mgt_schema_version + 1):
                        if sql_version > db_version:
                            path = "{}/python/foglamp/services/management/data/migration/up/{}.sql".format(
                                _FOGLAMP_ROOT, sql_version)
                            db.run_sql_file(db_conn, path)
                elif mgt_schema_version < db_version:
                    _logger.info("Management DB schema downgrading from [{}] to [{}]...".format(
                        db_version, mgt_schema_version))
                    """ If mgt_schema_version=3; db_version=5 then;
                        run sql files (4.sql, 3.sql) from path data/migration/down
                    """
                    for sql_version in range(mgt_schema_version, db_version):
                        path = "{}/python/foglamp/services/management/data/migration/down/{}.sql".format(_FOGLAMP_ROOT,
                                                                                                         sql_version)
                        db.run_sql_file(db_conn, path)
                else:
                    _logger.info("Management DB schema is up to date with version [{}]".format(mgt_schema_version))
                # update DB version id with current management schema version
                query = "UPDATE version SET id = '{}'".format(mgt_schema_version)
                db.update_table(db_conn, query)
        else:
            db.load_database(mgt_schema_version)
            _logger.info("SQLite 3 database '{}' created.".format(db._DB_PATH))

        # Find core service port
        jdoc = cls.make_http_request('GET', '/foglamp/service/category/rest_api', payload=None)
        port_from_config = jdoc['httpPort']['value']
        if jdoc["enableHttp"]["value"] == "false":
            port_from_config = jdoc['httpsPort']['value']
            cls._is_tls_enabled = True
        cls._core_service_port = int(port_from_config)

        # Find host IP address & host ID
        cls._host, host_id = utils.get_host_info()

        # Starts the aiohttp server with management port
        cls.management_app = web.Application()
        management.setup(cls.management_app)
        cls.loop = asyncio.get_event_loop()
        cls.management_app_server, cls.management_app_server_handler = cls.start_app(cls.management_app,
                                                                                     cls._host, cls._management_port)
        _logger.info("Management Service server with management port started...")
        cls._management_port = cls.management_app_server.sockets[0].getsockname()[1]

        # Starts the aiohttp server with service port
        cls.service_app = web.Application()
        service.setup(cls.service_app)
        cls.service_app_server, cls.service_app_server_handler = cls.start_app(cls.service_app, cls._host,
                                                                               cls._service_port)
        _logger.info("Management Service server with service port started...")
        cls._service_port = cls.service_app_server.sockets[0].getsockname()[1]

        # TODO: protocol=http/https - dynamic
        service_registration_payload = {
            "name": cls._service_name,
            "type": "Management",
            "management_port": cls._management_port,
            "service_port": cls._service_port,
            "address": cls._host,
            "protocol": "http"
        }
        result = cls.register_service(service_registration_payload, registration_token)
        _logger.debug("Result for service registration: {}".format(result))
        b_token = result['bearer_token']  # get result bearer token
        # if registraton fails/token or other issue,
        # log error, shut down the started servers & exit
        b_token = "Bearer " + b_token
        # Find authToken, when authentication is optional then ignore otherwise set authentication token
        jdoc = cls.make_http_request('GET', '/foglamp/service/authtoken', payload=None, port=None,
                                     headers={"Authorization": b_token})
        if 'token' in jdoc:
            _logger.info("FogLAMP authentication is mandatory.")
            cls._auth_token = jdoc['token']
        elif 'message' in jdoc:
            # HTTPPreconditionFailed tells auth is optional
            if jdoc['message'] == "Precondition Failed":
                pass
            else:
                _logger.error(jdoc['message'])
                return
        else:
            _logger.error("Unknown error %s", jdoc)
            return

        result = cls.create_category()
        _logger.debug("Result for category creation: {}".format(result))
        result = cls.register_interest()
        _logger.debug("Result for interest registration: {}".format(result))
        result = cls.add_schedule_for_task(host_id)
        _logger.debug("Result for schedule task: {}".format(result))
        cls.loop.run_forever()
